"""A tiled image gadget."""

from math import ceil

from ..geometry import Region
from .gadget import Gadget, Point, Pointlike, PosHint, Size, SizeHint, Sizelike

__all__ = ["Point", "Size", "Tiled"]


class Tiled(Gadget):
    r"""
    Tile a gadget over the visible region of ``Tiled``.

    Parameters
    ----------
    tile : Gadget
        The gadget to tile.
    size : Sizelike, default: Size(10, 10)
        Size of gadget.
    pos : Pointlike, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: True
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    tile : Graphics
        The gadget to tile.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : TotalSizeHint
        Size as a proportion of parent's height and width.
    pos_hint : TotalPosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App | None
        The running app.

    Methods
    -------
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(gadget_it, \*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        tile: Gadget,
        size: Sizelike = Size(10, 10),
        pos: Pointlike = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = True,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        super().__init__(
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        self.tile = tile

    def _render(self, cells, graphics, kind):
        # ! Maybe add custom renderer?
        h, w = self._size
        th, tw = tsize = self.tile.size
        oy, ox = self.absolute_pos
        for y in range(ceil(h / th)):
            for x in range(ceil(w / tw)):
                self.tile.pos = oy + y * th, ox + x * tw
                self.tile._region = (
                    Region.from_rect(self.tile.pos, tsize) & self._region
                )
                self.tile._render(cells, graphics, kind)



"""A beam effect."""

import asyncio
from collections.abc import Callable
from random import random, shuffle
from typing import Coroutine, Literal

import numpy as np

from ...array_types import Int2D
from ...colors import WHITE, Color, gradient
from ..text import Text

HORIZONTAL_BEAM = "▁▁▁▁▁▁▁▁▂▂▂▂▂▂▂▂▃▃▃▃▃▃▃▃▄▄▄▄▄▄▄▄"
VERTICAL_BEAM = "▏\n▏\n▏\n▏\n▎\n▎\n▎\n▎\n▍\n▍\n▍\n▍\n▌\n▌\n▌\n▌"
BEAM_BLUE = Color.from_hex("00d1ff")
BEAM_PURP = Color.from_hex("8a008a")
HORIZONTAL_GRAD = gradient(BEAM_PURP, BEAM_BLUE, WHITE, n=32)
VERTICAL_GRAD = gradient(BEAM_PURP, BEAM_BLUE, WHITE, n=16)


async def beams_effect(text: Text):
    """
    Create a beams effect.

    Parameters
    ----------
    text : Text
        The text gadget which to run the effect.

    Warnings
    --------
    Modifying `text` size while effect is running will break the effect.
    """
    cover = Text(size=text.size)
    cover.canvas[:] = text.canvas
    cover.canvas["fg_color"] = cover.canvas["bg_color"]

    pass_count = np.zeros(text.size, int)
    beams, tweens = _create_beams(pass_count, cover)
    shuffle(tweens)

    cover.add_gadgets(beams)
    text.add_gadget(cover)

    tasks = [asyncio.create_task(_fade_cover(pass_count, cover, text))]
    for tween in tweens:
        tasks.append(asyncio.create_task(tween))
        await asyncio.sleep(0.1 * random())

    await asyncio.gather(*tasks)

    text.remove_gadget(cover)


def _create_on_progress(
    pass_count: Int2D,
    cover: Text,
    beam: Text,
    kind: Literal["down", "left", "right", "up"],
) -> Callable[[float], None]:
    if kind == "down":
        last_bottom = beam.bottom

        def on_progress(_):
            nonlocal last_bottom
            if beam.bottom != last_bottom:
                pass_count[last_bottom + 1 : beam.bottom + 1, beam.x] += 1
                cover.canvas[last_bottom + 1 : beam.bottom + 1, beam.x]["fg_color"] = (
                    WHITE
                )
                last_bottom = beam.bottom

    elif kind == "left":
        last_left = beam.left

        def on_progress(_):
            nonlocal last_left
            if beam.left != last_left:
                left = max(beam.left, 0)
                pass_count[beam.y, left:last_left] += 1
                cover.canvas[beam.y, left:last_left]["fg_color"] = WHITE
                last_left = beam.left

    elif kind == "right":
        last_right = beam.right

        def on_progress(_):
            nonlocal last_right
            if beam.right != last_right:
                pass_count[beam.y, last_right + 1 : beam.right + 1] += 1
                cover.canvas[beam.y, last_right + 1 : beam.right + 1]["fg_color"] = (
                    WHITE
                )
                last_right = beam.right

    else:
        last_top = beam.top

        def on_progress(_):
            nonlocal last_top
            if beam.top != last_top:
                top = max(beam.top, 0)
                pass_count[top:last_top, beam.x] += 1
                cover.canvas[top:last_top, beam.x]["fg_color"] = WHITE
                last_top = beam.top

    return on_progress


def _create_beams(pass_count: Int2D, cover: Text) -> tuple[list[Text], list[Coroutine]]:
    h, w = cover.size
    beams = []
    tweens = []
    for y in range(h):
        beam = Text(is_transparent=True)
        beam.y = y
        duration = 0.3 + random()
        if round(random()):
            beam.set_text(HORIZONTAL_BEAM)
            beam.canvas["fg_color"] = HORIZONTAL_GRAD
            beam.right = -1
            tween = beam.tween(
                duration=duration,
                on_progress=_create_on_progress(pass_count, cover, beam, "right"),
                left=w,
            )
        else:
            beam.set_text(HORIZONTAL_BEAM[::-1])
            beam.canvas["fg_color"] = HORIZONTAL_GRAD[::-1]
            beam.left = w
            tween = beam.tween(
                duration=duration,
                on_progress=_create_on_progress(pass_count, cover, beam, "left"),
                right=-1,
            )
        beams.append(beam)
        tweens.append(tween)

    for x in range(w):
        beam = Text(is_transparent=True)
        beam.x = x
        duration = 0.3 + random()
        if round(random()):
            beam.set_text(VERTICAL_BEAM)
            beam.canvas["fg_color"][:, 0] = VERTICAL_GRAD
            beam.bottom = -1
            tween = beam.tween(
                duration=duration,
                on_progress=_create_on_progress(pass_count, cover, beam, "down"),
                top=h,
            )
        else:
            beam.set_text(VERTICAL_BEAM[::-1])
            beam.canvas["fg_color"][:, 0] = VERTICAL_GRAD[::-1]
            beam.top = h
            tween = beam.tween(
                duration=duration,
                on_progress=_create_on_progress(pass_count, cover, beam, "up"),
                bottom=-1,
            )
        beams.append(beam)
        tweens.append(tween)
    return beams, tweens


async def _fade_cover(pass_count: Int2D, cover: Text, text: Text):
    cover_fg = cover.canvas["fg_color"]
    text_fg = text.canvas["fg_color"].astype(float)  # type: ignore
    while not np.array_equal(cover_fg, text_fg):
        passed_mask = pass_count >= 1
        faded = cover_fg * 0.99 + text_fg * 0.01
        cover_fg[passed_mask] = faded.astype(np.uint8)[passed_mask]
        await asyncio.sleep(0)



"""An animated toggle button gadget."""

import asyncio
from collections.abc import Callable, Hashable

from ..colors import GREEN, Color
from ..text_tools import Cell0D, Style, smooth_horizontal_bar
from .behaviors.toggle_button_behavior import ToggleButtonBehavior, ToggleState
from .text import Point, Pointlike, PosHint, Size, SizeHint, Sizelike, Text

__all__ = ["FlatToggle", "Point", "Size"]

DARK_GREY = Color.from_hex("222222")
LIGHT_GREY = Color.from_hex("666666")
DARK_RED = Color.from_hex("4f0908")
DARK_GREEN = Color.from_hex("0e4f08")


class FlatToggle(ToggleButtonBehavior, Text):
    r"""
    An animated toggle button gadget.

    Parameters
    ----------
    callback : Callable[ToggleState], None] | None, default: None
        Called when button is released.
    group : Hashable | None, default: None
        If a group is provided, only one button in a group can be in the on state.
    allow_no_selection : bool, default: False
        If a group is provided, setting this to true allows no selection, i.e.,
        every button can be in the off state.
    always_release : bool, default: False
        Whether a mouse up event outside the button will trigger it.
    default_cell : Cell0D | str, default: " "
        Default cell of text canvas.
    alpha : float, default: 0.0
        Transparency of gadget.
    size : Sizelike, default: Size(1, 3)
        Size of gadget.
    pos : Pointlike, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    callback : Callable[[ToggleState], None] | None
        Called when button is released.
    group : Hashable | None
        If a group is provided, only one button in a group can be in the on state.
    allow_no_selection : bool
        If true and button is in a group, every button can be in the off state.
    toggle_state : ToggleState
        Toggle state of button.
    always_release : bool
        Whether a mouse up event outside the button will trigger it.
    button_state : ButtonState
        Current button state.
    canvas : Cell2D
        The array of characters for the gadget.
    chars : Unicode2D
        Return a view of the ords field of the canvas as 1-character unicode strings.
    default_cell : Cell0D
        Default cell of text canvas.
    default_fg_color : Color
        Foreground color of default cell.
    default_bg_color : Color
        Background color of default cell.
    alpha : float
        Transparency of gadget.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : TotalSizeHint
        Size as a proportion of parent's height and width.
    pos_hint : TotalPosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App | None
        The running app.

    Methods
    -------
    on_toggle()
        Triggled on toggle state change.
    update_off()
        Paint the off state.
    update_on()
        Paint the on state.
    on_release()
        Triggered when a button is released.
    update_normal()
        Paint the normal state.
    update_hover()
        Paint the hover state.
    update_down()
        Paint the down state.
    update_disallowed()
        Paint the disallowed state.
    add_border(style="light", ...)
        Add a border to the gadget.
    add_syntax_highlighting(lexer, style)
        Add syntax highlighting to current text in canvas.
    add_str(str, pos, ...)
        Add a single line of text to the canvas.
    set_text(text, ...)
        Resize gadget to fit text, erase canvas, then fill canvas with text.
    clear()
        Fill canvas with default cell.
    shift(n=1)
        Shift content in canvas up (or down in case of negative `n`).
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(gadget_it, \*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        callback: Callable[[ToggleState], None] | None = None,
        group: None | Hashable = None,
        allow_no_selection: bool = False,
        always_release: bool = False,
        default_cell: Cell0D | str = " ",
        alpha: float = 0.0,
        size: Sizelike = Size(1, 3),
        pos: Pointlike = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        super().__init__(
            group=group,
            allow_no_selection=allow_no_selection,
            always_release=always_release,
            default_cell=default_cell,
            alpha=alpha,
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        self.callback = callback
        """Called when button is released."""
        self._animation_task: asyncio.Task | None = None
        self._animation_progress: float
        self._animation_progress = float(self.toggle_state == "on")
        self._draw_toggle()

    def _update_color(self):
        if self.button_state == "disallowed":
            on_color = DARK_GREEN
            off_color = DARK_RED
        else:
            on_color = GREEN
            off_color = LIGHT_GREY

        self.canvas["fg_color"] = DARK_GREY
        self.canvas["bg_color"] = on_color if self.toggle_state == "on" else off_color

    def update_off(self) -> None:
        """Paint the off state."""
        self._update_color()

    def update_on(self) -> None:
        """Paint the on state."""
        self._update_color()

    def update_normal(self) -> None:
        """Paint the normal state."""
        self._update_color()

    def update_down(self) -> None:
        """Paint the down state."""
        self._update_color()

    def update_hover(self) -> None:
        """Paint the hover state."""
        self._update_color()

    def update_disallowed(self) -> None:
        """Paint the disallowed state."""
        self._update_color()

    def _draw_toggle(self, _=0.0) -> None:
        self.clear()
        self._update_color()

        if self.width < 2:
            return

        x, p = divmod((self.width - 1.25) * self._animation_progress + 0.125, 1)
        x = int(x)
        bar = smooth_horizontal_bar(1, 1, p)
        self.chars[:, x : x + 2] = bar
        self.canvas["style"][:, x] = Style.REVERSE

    def on_toggle(self) -> None:
        """Animate toggle."""
        if self._animation_task is not None:
            self._animation_task.cancel()
        if self.callback is not None:
            self.callback(self.toggle_state)

        p = float(self.toggle_state == "on")
        self._animation_task = asyncio.create_task(
            self.tween(
                duration=abs(self._animation_progress - p) * 0.3,
                on_progress=self._draw_toggle,
                easing="in_quint",
                _animation_progress=p,
            )
        )

    def on_size(self) -> None:
        """Redraw toggle on resize."""
        super().on_size()
        self._draw_toggle()

    def on_remove(self) -> None:
        """Stop animation on remove."""
        if self._animation_task is not None:
            self._animation_task.cancel()
        super().on_remove()



"""A textbox gadget for single-line editable text."""

from __future__ import annotations

from collections.abc import Callable
from dataclasses import astuple, dataclass
from typing import Literal, cast

from ugrapheme import graphemes
from uwcwidth import wcswidth

from ..geometry import rect_slice
from ..logging import get_logger
from ..terminal.events import KeyEvent, MouseButton, MouseEvent, PasteEvent
from ..text_tools import canvas_as_text, egc_chr, egc_ord, is_word_char
from .behaviors.focusable import Focusable
from .behaviors.grabbable import Grabbable
from .behaviors.themable import Themable
from .cursor import Cursor
from .gadget import (
    Gadget,
    Point,
    Pointlike,
    PosHint,
    Region,
    Size,
    SizeHint,
    Sizelike,
)
from .text import Text

__all__ = ["Point", "Size", "Textbox"]

logger = get_logger(__name__)


@dataclass
class _TextSelection:
    start: int
    end: int


@dataclass
class _TextEdit:
    cursor: int
    selection: _TextSelection | None
    start: int
    end: int
    text: str

    @property
    def first(self) -> int:
        return min(self.start, self.end)

    @first.setter
    def first(self, first: int):
        if self.start <= self.end:
            self.start = first
        else:
            self.end = first

    @property
    def last(self) -> int:
        return max(self.start, self.end)

    @last.setter
    def last(self, last: int):
        if self.end >= self.start:
            self.end = last
        else:
            self.start = last

    def end_of_content(self) -> int:
        return self.first + wcswidth(self.text)

    def join(self, other: _TextEdit) -> bool:
        if other.end_of_content() == self.first:
            width = self.last - self.first
            self.first = other.first
            self.last = other.last + width
            self.text = other.text + self.text
            return True

        if self.last == other.first:
            self.last = other.last
            self.text += other.text
            return True

        return False


class _Box(Text):
    def _render(self, cells, graphics, kind):
        super()._render(cells, graphics, kind)
        textbox = cast(Textbox, self.parent)
        if textbox.hide_input:
            hider_rect = Region.from_rect(self.absolute_pos, (1, textbox._line_width))
            hider_region = self._region & hider_rect
            for pos, size in hider_region.rects():
                cells["ord"][rect_slice(pos, size)] = textbox._hide_char_ord


class Textbox(Themable, Focusable, Grabbable, Gadget):
    r"""
    A textbox gadget for single-line editable text.

    Supports pasting, mouse selection, and cursor navigation.

    Parameters
    ----------
    enter_callback : Callable[[Textbox], None] | None, default: None
        Called when textbox has focus and `enter` is pressed.
    placeholder : str, default: ""
        Placeholder text for textbox.
    hide_input : bool, default: False
        Whether input is hidden with :attr:`hide_char`.
    hide_char : str, default: "*"
        Character to hide input when :attr:`hide_input` is true.
    max_chars : int | None, default: None
        Maximum allowed number of characters in textbox.
    alpha : float, default: 1.0
        Transparency of gadget.
    is_grabbable : bool, default: True
        Whether grabbable behavior is enabled.
    ptf_on_grab : bool, default: False
        Whether the gadget will be pulled to front when grabbed.
    mouse_button : MouseButton | Literal["any"], default: "left"
        Mouse button used for grabbing.
    size : Sizelike, default: Size(10, 10)
        Size of gadget.
    pos : Pointlike, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    enter_callback : Callable[[Textbox], None] | None
        Called when textbox has focus and `enter` is pressed.
    placeholder : str
        Placeholder text for textbox.
    hide_input : bool
        Whether input is hidden with :attr:`hide_char`.
    hide_char : str
        Character to hide input when :attr:`hide_input` is true.
    max_chars : int | None
        Maximum allowed number of characters in textbox.
    alpha : float
        Transparency of gadget.
    text : str
        The textbox's text.
    cursor : int
        The cursor column.
    is_selecting : bool
        Whether there is a selection.
    has_nonempty_selection : bool
        Whether selection is non-empty.
    is_focused : bool
        Whether gadget has focus.
    any_focused : bool
        Whether any gadget has focus.
    is_grabbable : bool
        Whether grabbable behavior is enabled.
    ptf_on_grab : bool
        Whether the gadget will be pulled to front when grabbed.
    mouse_button : MouseButton | Literal["any"]
        Mouse button used for grabbing.
    is_grabbed : bool
        Whether gadget is grabbed.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : TotalSizeHint
        Size as a proportion of parent's height and width.
    pos_hint : TotalPosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App | None
        The running app.

    Methods
    -------
    replace_text(start, end, content)
        Replace text from ``start`` to ``end`` with ``content``.
    undo()
        Undo previous edit.
    redo()
        Redo previous undo.
    move_cursor_left(n)
        Move cursor left `n` characters.
    move_cursor_right(n)
        Move cursor right `n` characters.
    move_word_left()
        Move cursor a word left.
    move_word_right()
        Move cursor a word right.
    get_color(color_name)
        Get a color by name from the current color theme.
    update_theme()
        Paint the gadget with current theme.
    focus()
        Focus gadget.
    blur()
        Un-focus gadget.
    focus_next()
        Focus next focusable gadget.
    focus_previous()
        Focus previous focusable gadget.
    on_focus()
        Update gadget when it gains focus.
    on_blur()
        Update gadget when it loses focus.
    grab(mouse_event)
        Grab the gadget.
    ungrab(mouse_event)
        Ungrab the gadget.
    grab_update(mouse_event)
        Update gadget with incoming mouse events while grabbed.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(gadget_it, \*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        enter_callback: Callable[[Textbox], None] | None = None,
        placeholder: str = "",
        hide_input: bool = False,
        hide_char: str = "*",
        max_chars: int | None = None,
        is_grabbable: bool = True,
        ptf_on_grab: bool = False,
        mouse_button: MouseButton | Literal["any"] = "left",
        alpha: float = 1.0,
        size: Sizelike = Size(10, 10),
        pos: Pointlike = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        self._placeholder_gadget = Text(alpha=0.0, is_transparent=is_transparent)
        self._placeholder_gadget.set_text(placeholder)
        self._cursor = Cursor(reverse=False)
        self._box = _Box(size=size, is_transparent=is_transparent)
        super().__init__(
            is_grabbable=is_grabbable,
            ptf_on_grab=ptf_on_grab,
            mouse_button=mouse_button,
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )

        self._selection: _TextSelection | None = None
        self._line_width = 0
        self._replace_text_to_undo_stack: bool = True
        """Whether ``replace_text`` uses the undo or redo stack."""
        self._replace_text_clears_redo: bool = True
        """Whether ``replace_text`` clears the redo stack."""
        self._undo_stack: list[_TextEdit] = []
        self._redo_stack: list[_TextEdit] = []

        self._box.add_gadgets(self._placeholder_gadget, self._cursor)
        self.add_gadgets(self._box)

        self.enter_callback = enter_callback
        """Called when textbox has focus and `enter` is pressed."""
        self.placeholder = placeholder
        """Placeholder text for textbox."""
        self.hide_input = hide_input
        """Whether input is hidden with :attr:`hide_char`."""
        self._hide_char_ord: int = 0
        """``hide_char``'s ordinal representation."""
        self.hide_char = hide_char
        """Character to hide input if :attr:`hide_input` is true."""
        self.max_chars = max_chars
        """Maximum allowed number of characters in textbox."""
        self.alpha = alpha

    @property
    def alpha(self) -> float:
        """Transparency of gadget."""
        return self._box.alpha

    @alpha.setter
    def alpha(self, alpha: float) -> None:
        self._box.alpha = alpha

    @property
    def hide_char(self) -> str:
        """
        Character to hide input when :attr:`hide_input` is true.

        If ``hide_char`` cell width is not 1 it will default to ``"*"``.
        """
        return egc_chr(self._hide_char_ord)

    @hide_char.setter
    def hide_char(self, char: str) -> None:
        self._hide_char_ord = egc_ord(char)
        if wcswidth(egc_chr(self._hide_char_ord)) != 1:
            logger.info(
                "hide_char (%s) cell width greater than 1, using default hide_char",
                char,
            )
            self._hide_char_ord = ord("*")

    @property
    def placeholder(self) -> str:
        """Placeholder text for textbox."""
        return self._placeholder

    @placeholder.setter
    def placeholder(self, placeholder: str):
        self._placeholder = placeholder
        self._placeholder_gadget.set_text(placeholder)
        self._placeholder_gadget.is_enabled = self._line_width == 0 and bool(
            placeholder
        )

    @property
    def text(self) -> str:
        """The textbox's text."""
        return canvas_as_text(self._box.canvas[0, : self._line_width])

    @text.setter
    def text(self, text: str) -> None:
        self._selection = None
        text = graphemes(text.replace("\n", " "))[: self.max_chars]
        self.replace_text(0, self._line_width, text)
        self._undo_stack.clear()
        self._redo_stack.clear()
        self.cursor = self._line_width

    @property
    def cursor(self) -> int:
        """The cursor column."""
        return self._cursor.x

    @cursor.setter
    def cursor(self, cursor: int):
        """After setting cursor position, move textbox so that cursor is visible."""
        self._cursor.x = cursor
        self._placeholder_gadget.is_enabled = self._line_width == 0 and bool(
            self._placeholder
        )

        max_x = self.width - 1
        rel_x = cursor + self._box.x
        if rel_x > max_x:
            self._box.x -= rel_x - max_x
        elif rel_x < 0:
            self._box.x -= rel_x

        if self._selection is not None:
            self._selection.end = self.cursor
        self._highlight_selection()

    def replace_text(self, start: int, end: int, content: str) -> None:
        """
        Replace text from ``start`` to ``end`` with ``content``.

        Parameters
        ----------
        start : int
            The start of text to replace.
        end : int
            The end of text to replace.
        content : str
            The replacement text.
        """
        if start > end:
            start, end = end, start

        if end > self._line_width:
            logger.debug("End of delete greater than line length.")
            end = self._line_width

        prev_selection = self._selection
        prev_cursor = self.cursor
        text = canvas_as_text(self._box.canvas[0, start:end])

        self._selection = None

        len_end = self._line_width - end
        line_width = start + len_end

        box = self._box
        box.canvas[0, start:line_width] = box.canvas[0, end : end + len_end]
        box.canvas[0, line_width:] = box.default_cell

        content = content.replace("\n", " ")

        box_text = graphemes(
            canvas_as_text(box.canvas[0, :start])
            + content
            + canvas_as_text(box.canvas[0, start:line_width])
        )[: self.max_chars]

        box_width = self._line_width = wcswidth(box_text)
        if box_width >= box.width:
            box.width = box_width + 1

        box.add_str(box_text)
        box.canvas[0, box_width:] = box.default_cell

        self.cursor = start + wcswidth(content)

        inverse_edit = _TextEdit(prev_cursor, prev_selection, start, self.cursor, text)
        if self._replace_text_to_undo_stack:
            stack = self._undo_stack
        else:
            stack = self._redo_stack
            self._replace_text_to_undo_stack = True

        if self._replace_text_clears_redo:
            self._redo_stack.clear()
        else:
            self._replace_text_clears_redo = True

        if not (stack and stack[-1].join(inverse_edit)):
            stack.append(inverse_edit)

    def undo(self):
        """Undo previous edit."""
        if self._undo_stack:
            last_edit = self._undo_stack.pop()
            self._replace_text_to_undo_stack = False
            self._replace_text_clears_redo = False
            self.replace_text(last_edit.start, last_edit.end, last_edit.text)
            self._selection = last_edit.selection
            self.cursor = last_edit.cursor

    def redo(self):
        """Redo previous undo."""
        if self._redo_stack:
            prev_edit = self._redo_stack.pop()
            self._replace_text_clears_redo = False
            self.replace_text(prev_edit.start, prev_edit.end, prev_edit.text)
            self._selection = prev_edit.selection
            self.cursor = prev_edit.cursor

    def _highlight_selection(self):
        fg = self._box.canvas["fg_color"]
        bg = self._box.canvas["bg_color"]
        fg[:] = self._box.default_fg_color
        bg[:] = self._box.default_bg_color
        if self._selection is not None and self._selection.start != self._selection.end:
            start, end = self._selection.start, self._selection.end
            if start > end:
                start, end = end, start
            fg[0, start:end] = self.get_color("textbox_selection_highlight_fg")
            bg[0, start:end] = self.get_color("textbox_selection_highlight_bg")

    def move_cursor_left(self, n: int = 1):
        """Move cursor left `n` characters."""
        text_before_cursor = canvas_as_text(self._box.canvas[0, : self.cursor])
        egcs = graphemes(text_before_cursor)
        if n <= len(egcs):
            self.cursor = wcswidth(egcs[:-n])
        else:
            self.cursor = 0

    def move_cursor_right(self, n: int = 1):
        """Move cursor right `n` characters."""
        text_after_cursor = canvas_as_text(
            self._box.canvas[0, self.cursor : self._line_width]
        )
        egcs = graphemes(text_after_cursor)
        if n <= len(egcs):
            self.cursor += wcswidth(egcs[:n])
        else:
            self.cursor = self._line_width

    def move_word_left(self):
        """Move cursor a word left."""
        last_x = self.cursor
        first_char_found = char_is_word_char = False
        while True:
            self.move_cursor_left()
            if self.cursor == last_x:
                break

            last_x = self.cursor

            current_char = egc_chr(self._box.canvas[0, self.cursor]["ord"])
            if not first_char_found:
                if not current_char.isspace():
                    first_char_found = True
                    char_is_word_char = is_word_char(current_char)
            elif current_char.isspace() or char_is_word_char != is_word_char(
                current_char
            ):
                self.move_cursor_right()
                break

    def move_word_right(self):
        """Move cursor a word right."""
        last_x = self.cursor
        first_char_found = char_is_word_char = False
        while True:
            self.move_cursor_right()
            if self.cursor == last_x:
                break

            last_x = self.cursor

            current_char = egc_chr(self._box.canvas[0, self.cursor]["ord"])
            if not first_char_found:
                if not current_char.isspace():
                    first_char_found = True
                    char_is_word_char = is_word_char(current_char)
            elif current_char.isspace() or char_is_word_char != is_word_char(
                current_char
            ):
                break

    def _enter(self):
        if self.enter_callback is not None:
            self.enter_callback(self)

    def _backspace(self):
        if self._selection is None:
            end = self.cursor
            self.move_cursor_left()
            start = self.cursor
            self.cursor = end
            if start == end:
                return
        else:
            start, end = self._selection.start, self._selection.end
        self.replace_text(start, end, "")

    def _delete(self):
        if self._selection is None:
            start = self.cursor
            self.move_cursor_right()
            end = self.cursor
            self.cursor = start
            if start == end:
                return
        else:
            start, end = self._selection.start, self._selection.end
        self.replace_text(start, end, "")

    def _left(self):
        if self._selection is None:
            self.move_cursor_left()
        else:
            select_start = min(self._selection.start, self._selection.end)
            self._selection = None
            self.cursor = select_start

    def _right(self):
        if self._selection is None:
            self.move_cursor_right()
        else:
            select_end = max(self._selection.start, self._selection.end)
            self._selection = None
            self.cursor = select_end

    def _ctrl_left(self):
        self._selection = None
        self.move_word_left()

    def _ctrl_right(self):
        self._selection = None
        self.move_word_right()

    def _ctrl_a(self):
        """Select all."""
        self._selection = _TextSelection(0, self._line_width)
        self.cursor = self._line_width

    def _ctrl_d(self):
        """Select word."""
        self._selection = None
        last_x = self.cursor
        while True:
            self.move_cursor_left()
            if last_x == self.cursor:
                break
            if not is_word_char(egc_chr(self._box.canvas[0, self.cursor]["ord"])):
                self.move_cursor_right()
                break
            last_x = self.cursor

        self._selection = _TextSelection(self.cursor, self.cursor)
        last_x = self.cursor
        while True:
            if not is_word_char(egc_chr(self._box.canvas[0, self.cursor]["ord"])):
                break
            self.move_cursor_right()
            if last_x == self.cursor:
                break
            last_x = self.cursor

    def _home(self):
        self._selection = None
        self.cursor = 0

    def _end(self):
        self._selection = None
        self.cursor = self._line_width

    def _shift_left(self):
        if self._selection is None:
            self._selection = _TextSelection(self.cursor, self.cursor)
        self.move_cursor_left()

    def _shift_right(self):
        if self._selection is None:
            self._selection = _TextSelection(self.cursor, self.cursor)
        self.move_cursor_right()

    def _shift_ctrl_left(self):
        if self._selection is None:
            self._selection = _TextSelection(self.cursor, self.cursor)
        self.move_word_left()

    def _shift_ctrl_right(self):
        if self._selection is None:
            self._selection = _TextSelection(self.cursor, self.cursor)
        self.move_word_right()

    def _shift_home(self):
        if self._selection is None:
            self._selection = _TextSelection(self.cursor, self.cursor)
        self.cursor = 0

    def _shift_end(self):
        if self._selection is None:
            self._selection = _TextSelection(self.cursor, self.cursor)
        self.cursor = self._line_width

    def _escape(self):
        if self._selection is None:
            self.blur()
        else:
            self._selection = None
            self._highlight_selection()

    def _ascii(self, key: str):
        if self._selection is None:
            start = end = self.cursor
        else:
            start, end = self._selection.start, self._selection.end
        self.replace_text(start, end, key)

    __HANDLERS = {
        ("enter", False, False, False): _enter,
        ("backspace", False, False, False): _backspace,
        ("delete", False, False, False): _delete,
        ("left", False, False, False): _left,
        ("right", False, False, False): _right,
        ("left", False, True, False): _ctrl_left,
        ("right", False, True, False): _ctrl_right,
        ("home", False, False, False): _home,
        ("end", False, False, False): _end,
        ("left", False, False, True): _shift_left,
        ("right", False, False, True): _shift_right,
        ("left", False, True, True): _shift_ctrl_left,
        ("right", False, True, True): _shift_ctrl_right,
        ("home", False, False, True): _shift_home,
        ("end", False, False, True): _shift_end,
        ("escape", False, False, False): _escape,
        ("z", False, True, False): undo,
        ("z", False, True, True): redo,
        ("y", False, True, False): redo,  # ctrl + shift + z won't work on linux
        ("r", False, True, False): redo,
        ("a", False, True, False): _ctrl_a,
        ("d", False, True, False): _ctrl_d,
    }

    def on_key(self, key_event: KeyEvent) -> bool | None:
        """Process key press."""
        if not self.is_focused:
            return super().on_key(key_event)

        if (
            not key_event.alt
            and not key_event.ctrl
            and not key_event.shift
            and len(key_event.key) == 1
        ):
            self._ascii(key_event.key)
        elif handler := self.__HANDLERS.get(astuple(key_event)):
            handler(self)
        else:
            return super().on_key(key_event)

        return True

    def on_paste(self, paste_event: PasteEvent) -> bool | None:
        """Add paste to textbox."""
        if not self.is_focused:
            return

        if self._selection is None:
            start = end = self.cursor
        else:
            start, end = self._selection.start, self._selection.end
        self.replace_text(start, end, paste_event.paste)
        self._redo_stack.clear()

        return True

    def grab(self, mouse_event):
        """Start selection on grab."""
        if self._box.collides_point(mouse_event.pos):
            super().grab(mouse_event)

            _, x = self._box.to_local(mouse_event.pos)
            cursor = min(x, self._line_width)

            if not mouse_event.shift or self._selection is None:
                self._selection = _TextSelection(cursor, cursor)

            self.cursor = cursor

    def grab_update(self, mouse_event: MouseEvent):
        """Update selection on grab update."""
        if self._box.collides_point(mouse_event.pos):
            _, x = self._box.to_local(mouse_event.pos)
            self.cursor = min(x, self._line_width)
        else:
            _, x = self.to_local(mouse_event.pos)

            if x < 0:
                self.move_cursor_left()
            elif x >= self.width:
                self.move_cursor_right()

    def ungrab(self, mouse_event):
        """Clear an empty selection on ungrab."""
        super().ungrab(mouse_event)
        if self._selection is not None and self._selection.start == self._selection.end:
            self._selection = None

    def on_transparency(self) -> None:
        """Update gadget after transparency is enabled/disabled."""
        self._box.is_transparent = self.is_transparent
        self._placeholder_gadget.is_transparent = self.is_transparent

    def update_theme(self) -> None:
        """Paint the gadget with current theme."""
        primary_fg = self.get_color("textbox_primary_fg")
        primary_bg = self.get_color("textbox_primary_bg")
        self._box.canvas["fg_color"] = self._box.default_fg_color = primary_fg
        self._box.canvas["bg_color"] = self._box.default_bg_color = primary_bg
        self._cursor.fg_color = primary_bg
        self._cursor.bg_color = primary_fg

        placeholder_fg = self.get_color("textbox_placeholder_fg")
        placeholder_bg = self.get_color("textbox_placeholder_bg")
        self._placeholder_gadget.default_fg_color = placeholder_fg
        self._placeholder_gadget.default_bg_color = placeholder_bg
        self._placeholder_gadget.canvas["fg_color"] = placeholder_fg
        self._placeholder_gadget.canvas["bg_color"] = placeholder_bg

        self._highlight_selection()

    def on_size(self):
        """Resize and reposition children on resize."""
        self._box.width = max(self.width, self._line_width + 1)
        if self._box.x + self._line_width < self.width:
            self._box.x = min(0, self.width - self._line_width)
        self.cursor = self.cursor

    def on_focus(self):
        """Show cursor and select all text on focus."""
        self._cursor.is_enabled = True
        if self._line_width > 0:
            self._ctrl_a()

    def on_blur(self):
        """Hide cursor on blur."""
        self._cursor.is_enabled = False
        self._selection = None
        self._highlight_selection()



"""A slider gadget."""

from collections.abc import Callable
from typing import Literal

from ..colors import BLACK, WHITE, Color
from ..terminal.events import MouseButton, MouseEvent
from .behaviors.grabbable import Grabbable
from .pane import (
    Pane,
    Point,
    Pointlike,
    PosHint,
    Size,
    SizeHint,
    Sizelike,
    bindable,
    clamp,
)
from .text import Text

__all__ = ["Point", "Size", "Slider"]


class Slider(Grabbable, Pane):
    r"""
    A slider gadget.

    Parameters
    ----------
    min : float
        Minimum value of slider.
    max : float
        Maximum value of slider.
    start_value: float | None, default: None
        Start value of slider. If `None`, start value is :attr:`min`.
    callback : Callable | None, default: None
        Single argument callable called with new value of slider when slider is updated.
    handle_color : Color, default: WHITE
        Color of handle.
    slider_color: Color, default: WHITE
        Color of slider.
    fill_color: Color, default: WHITE
        Color of filled portion of slider.
    slider_enabled : bool, default: True
        Whether slider value can be changed.
    is_grabbable : bool, default: True
        Whether grabbable behavior is enabled.
    ptf_on_grab : bool, default: False
        Whether the gadget will be pulled to front when grabbed.
    mouse_button : MouseButton | Literal["any"], default: "left"
        Mouse button used for grabbing.
    bg_color : Color, default: BLACK
        Background color of gadget.
    alpha : float, default: 1.0
        Transparency of gadget.
    size : Sizelike, default: Size(10, 10)
        Size of gadget.
    pos : Pointlike, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    min : float
        Minimum value of slider.
    max : float
        Maximum value of slider.
    value : float
        Current value of slider.
    callback : Callable
        Single argument callable called with new value of slider when slider is updated.
    slider_color: Color
        Color of slider.
    fill_color: Color
        Color of filled portion of slider.
    fill_color: Color
        Color of filled portion of slider.
    slider_enabled : bool
        Whether slider value can be changed.
    proportion : float
        Current proportion of slider.
    is_grabbable : bool
        Whether grabbable behavior is enabled.
    ptf_on_grab : bool
        Whether the gadget will be pulled to front when grabbed.
    mouse_button : MouseButton | Literal["any"]
        Mouse button used for grabbing.
    bg_color : Color
        Background color of gadget.
    alpha : float
        Transparency of gadget.
    is_grabbed : bool
        Whether gadget is grabbed.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : TotalSizeHint
        Size as a proportion of parent's height and width.
    pos_hint : TotalPosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App | None
        The running app.

    Methods
    -------
    grab(mouse_event)
        Grab the gadget.
    ungrab(mouse_event)
        Ungrab the gadget.
    grab_update(mouse_event)
        Update gadget with incoming mouse events while grabbed.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(gadget_it, \*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        min: float,
        max: float,
        start_value: float | None = None,
        callback: Callable | None = None,
        handle_color: Color = WHITE,
        slider_color: Color = WHITE,
        fill_color: Color = WHITE,
        slider_enabled: bool = True,
        is_grabbable: bool = True,
        ptf_on_grab: bool = False,
        mouse_button: MouseButton | Literal["any"] = "left",
        bg_color: Color = BLACK,
        alpha: float = 1.0,
        size: Sizelike = Size(10, 10),
        pos: Pointlike = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        super().__init__(
            is_grabbable=is_grabbable,
            ptf_on_grab=ptf_on_grab,
            mouse_button=mouse_button,
            bg_color=bg_color,
            alpha=alpha,
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )

        if min >= max:
            raise ValueError(f"{min=} >= {max=}")
        self._min = min
        self._max = max

        self._slider = Text(
            default_cell="━",
            size=(1, 1),
            pos_hint={"y_hint": 0.5, "anchor": "top"},
            size_hint={"width_hint": 1.0},
            is_transparent=True,
        )
        self._handle = Text(
            default_cell="█", size=(1, 1), pos_hint={"y_hint": 0.5, "anchor": "top"}
        )
        self.add_gadgets(self._slider, self._handle)

        self.handle_color = handle_color
        self.slider_color = slider_color
        self.fill_color = fill_color
        self.callback = callback
        self.slider_enabled = slider_enabled
        self.value = self.min if start_value is None else start_value

    @property
    def min(self) -> float:
        """Minimum value of slider."""
        return self._min

    @min.setter
    def min(self, value: float):
        if value >= self.max:
            raise ValueError("Min can't be greater than or equal to max.")
        self._min = value
        self.proportion = self.proportion

    @property
    def max(self) -> float:
        """Maximum value of slider."""
        return self._max

    @max.setter
    def max(self, value: float):
        if value <= self.min:
            raise ValueError("Max can't be less than or equal to min.")
        self._max = value
        self.proportion = self.proportion

    @property
    def handle_color(self) -> Color:
        """Color of handle."""
        return self._handle.default_fg_color

    @handle_color.setter
    def handle_color(self, handle_color: Color):
        self._handle.default_fg_color = handle_color
        self._handle.canvas["fg_color"] = handle_color

    @property
    def slider_color(self) -> Color:
        """Color of slider."""
        return self._slider.default_fg_color

    @slider_color.setter
    def slider_color(self, slider_color: Color):
        self._slider.default_fg_color = slider_color
        self._slider.canvas["fg_color"][:, self._handle.x :] = slider_color

    @property
    def fill_color(self) -> Color:
        """Color of "filled" portion of slider."""
        return self._fill_color

    @fill_color.setter
    def fill_color(self, color: Color):
        self._fill_color = color
        self._slider.canvas["fg_color"][:, : self._handle.x] = color

    @property
    def proportion(self) -> float:
        """Current proportion of slider."""
        return self._proportion

    @proportion.setter
    @bindable
    def proportion(self, value: float):
        if not self.slider_enabled:
            return

        self._proportion = clamp(value, 0, 1)
        self._value = (self.max - self.min) * self._proportion + self.min

        self._handle.x = x = round(self._proportion * self.fill_width)
        self._slider.canvas["fg_color"][:, :x] = self.fill_color
        self._slider.canvas["fg_color"][:, x:] = self.slider_color

        if self.callback is not None:
            self.callback(self._value)

    @property
    def value(self) -> float:
        """Current value of slider."""
        return self._value

    @value.setter
    @bindable
    def value(self, value: float):
        value = clamp(value, self.min, self.max)
        self.proportion = (value - self.min) / (self.max - self.min)

    @property
    def fill_width(self):
        """Width of the slider minus the width of the handle."""
        return self.width - self._handle.width

    def on_size(self):
        """Resize canvas and color arrays and reposition slider handle."""
        super().on_size()
        self.proportion = self.proportion

    def on_add(self):
        """Resize canvas and color arrays and reposition slider handle."""
        super().on_add()
        self.proportion = self.proportion

    def grab(self, mouse_event: MouseEvent):
        """Move handle to mouse position on grab."""
        if (
            self.collides_point(mouse_event.pos)
            and self.to_local(mouse_event.pos).y == self.height // 2
        ):
            super().grab(mouse_event)
            self.grab_update(mouse_event)

    def grab_update(self, mouse_event: MouseEvent):
        """Update proportion and handle position on grab update."""
        x = clamp(self.to_local(mouse_event.pos).x, 0, self.fill_width)
        self.proportion = 0 if self.fill_width == 0 else x / self.fill_width



"""
A graphic particle field.

A particle field specializes in rendering many single "pixel" children from an array of
particle positions and an array of particle colors.
"""

from typing import Any

import numpy as np

from .._rendering import graphics_field_render
from ..array_types import (
    RGBA_1D,
    RGBA_2D,
    RGBM_2D,
    Cell2D,
    Coords,
    Enum2D,
    braille_pixel_dtype,
)
from ..logging import get_logger
from .gadget import (
    Gadget,
    Point,
    Pointlike,
    PosHint,
    Size,
    SizeHint,
    Sizelike,
    bindable,
    clamp,
)
from .graphics import Blitter, Graphics, scale_geometry

__all__ = ["GraphicParticleField", "Point", "Size"]

logger = get_logger(__name__)


class GraphicParticleField(Gadget):
    r"""
    A graphic particle field.

    A particle field specializes in rendering many single "pixel" children from an array
    of particle positions and and an particle colors. Particles positions are a
    ``(N, 2)`` shaped array of floats. The decimal part of the positions allows
    different blitters to place correct "subpixel" characters. For instance a position
    ``(0.0, 0.0)`` might render the character ``"⠁"``, while ``(0.5, 0.0)`` might render
    ``"⠄"``.

    Parameters
    ----------
    particle_coords : Coords | None, default: None
        An array of particle positions with shape ``(N, 2)``.
    particle_colors : RGBA_1D | None, default: None
        A RGBA array of particle colors with shape ``(N, 4)``.
    particle_properties : dict[str, Any] | None, default: None
        Additional particle properties.
    alpha : float, default: 1.0
        Transparency of gadget.
    blitter : Blitter, default: "half"
        Determines how graphics are rendered.
    size : Sizelike, default: Size(10, 10)
        Size of gadget.
    pos : Pointlike, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: True
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    nparticles : int
        Number of particles in particle field.
    particle_coords : Coords
        An array of particle positions with shape `(N, 2)`.
    particle_colors : RGBA_1D
        A RGBA array of particle colors with shape `(N, 4)`.
    particle_properties : dict[str, Any]
        Additional particle properties.
    alpha : float
        Transparency of gadget.
    blitter : Blitter
        Determines how graphics are rendered.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : TotalSizeHint
        Size as a proportion of parent's height and width.
    pos_hint : TotalPosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App | None
        The running app.

    Methods
    -------
    particles_from_texture(texture)
        Set particle positions and colors from visible pixels of an RGBA texture.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(gadget_it, \*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        particle_coords: Coords | None = None,
        particle_colors: RGBA_1D | None = None,
        particle_properties: dict[str, Any] | None = None,
        alpha: float = 1.0,
        blitter: Blitter = "half",
        size: Sizelike = Size(10, 10),
        pos: Pointlike = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = True,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        super().__init__(
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        self.particle_coords: Coords
        """An array of particle positions with shape `(N, 2)`."""
        if particle_coords is None:
            self.particle_coords = np.zeros((0, 2), dtype=np.float64)  # type: ignore
        else:
            self.particle_coords = np.ascontiguousarray(
                particle_coords, dtype=np.float64
            )

        self.particle_colors: RGBA_1D
        """A RGBA array of particle colors with shape `(N, 4)`."""
        if particle_colors is None:
            self.particle_colors = np.zeros(
                (len(self.particle_coords), 4), dtype=np.uint8
            )  # type: ignore
        else:
            self.particle_colors = np.ascontiguousarray(particle_colors, dtype=np.uint8)

        self.particle_properties: dict[str, Any]
        """Additional particle properties."""
        if particle_properties is None:
            self.particle_properties = {}
        else:
            self.particle_properties = particle_properties

        self._braille_pixels = np.zeros(0, dtype=braille_pixel_dtype)
        """Array for building braille glyphs from particles."""

        self.alpha = alpha
        """Transparency of gadget."""
        self._blitter: Blitter
        self.blitter = blitter
        """Determines how graphics are rendered."""

    @property
    def alpha(self) -> float:
        """Transparency of gadget."""
        return self._alpha

    @alpha.setter
    @bindable
    def alpha(self, alpha: float):
        self._alpha = clamp(float(alpha), 0.0, 1.0)

    @property
    def blitter(self) -> Blitter:
        """Determines how graphics are rendered."""
        return self._blitter

    @blitter.setter
    def blitter(self, blitter: Blitter):
        if blitter not in Blitter.__args__:
            raise TypeError(f"{blitter} is not a valid blitter type.")
        if (
            blitter == "sixel"
            and not Graphics._sixel_support
            or blitter
            in {
                "octant",
                "quadrant",
                "sextant",
            }  # FIXME: Add rendering for block characters.
        ):
            self._blitter = "half"
            logger.info(
                "'%s' blitter not yet supported. Blitter set to 'half'.", blitter
            )
        else:
            self._blitter = blitter

        if blitter == "braille":
            self._braille_pixels = np.zeros(
                (self.height * self.width), dtype=braille_pixel_dtype
            )
        else:
            self._braille_pixels = np.zeros(0, dtype=braille_pixel_dtype)

    def on_add(self) -> None:
        """Change sixel blitter if sixel is not supported on add."""
        if self._blitter == "sixel" and not Graphics._sixel_support:
            self.blitter = "half"
        super().on_add()

    def on_size(self) -> None:
        """Resize braille pixels on size."""
        if self.blitter == "braille":
            self._braille_pixels = np.zeros(
                (self.height * self.width), dtype=braille_pixel_dtype
            )

    @property
    def nparticles(self) -> int:
        """Number of particles in particle field."""
        return len(self.particle_coords)

    def particles_from_texture(self, texture: RGBA_2D) -> None:
        """
        Set particle positions and colors from visible pixels of an RGBA texture.

        Parameters
        ----------
        texture : RGBA_2D
            A uint8 RGBA numpy array.
        """
        positions = np.argwhere(texture[..., 3])
        pys, pxs = positions.T
        self.particle_colors = np.ascontiguousarray(texture[pys, pxs])
        self.particle_coords = np.ascontiguousarray(positions.astype(np.float64))
        self.particle_coords /= scale_geometry(self._blitter, Size(1, 1))  # type: ignore

    def _render(
        self,
        cells: Cell2D,
        graphics: RGBM_2D,
        kind: Enum2D,
    ):
        """Render visible region of gadget."""
        graphics_field_render(
            cells,
            graphics,
            kind,
            self.absolute_pos,
            self._blitter,
            self._is_transparent,
            self.particle_coords,
            self.particle_colors,
            self._alpha,
            self._braille_pixels,
            self._region,
        )



from libc.stdlib cimport free, malloc, qsort
from libc.math cimport cos, exp, fabs, floor, sin

cimport cython

from .._rendering cimport Cell

ctypedef unsigned char uint8
ctypedef unsigned long ulong


cdef struct Sprite:
    double y
    double x
    size_t texture_idx
    double distance


cdef int sprite_cmp(const void* a, const void* b) noexcept nogil:
    cdef:
        const Sprite *sa = <Sprite*>a
        const Sprite *sb = <Sprite*>b
    if sa.distance > sb.distance:
        return -1
    if sa.distance == sb.distance:
        return 0
    return 1


cdef inline void composite(uint8 *dst, uint8 *src):
    if src[3] == 0:
        return

    if src[3] == 255 or dst[3] == 0:
        # Fast path for masking.
        dst[0] = src[0]
        dst[1] = src[1]
        dst[2] = src[2]
        dst[3] = src[3]
        return

    cdef double b = <double>dst[0], alpha = <double>src[3] / 255.0
    dst[0] = <uint8>((<double>src[0] - b) * alpha + b)
    b = <double>dst[1]
    dst[1] = <uint8>((<double>src[1] - b) * alpha + b)
    b = <double>dst[2]
    dst[2] = <uint8>((<double>src[2] - b) * alpha + b)
    dst[3] = src[3] + <uint8>(<double>dst[3] * (1 - alpha))


@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline void draw_floor_ceiling(
    uint8[:, :, ::1] texture,
    bint draw_ceil,
    uint8[:, :, ::1] ceiling_texture,
    bint draw_floor,
    uint8[:, :, ::1] floor_texture,
    double pos_y,
    double pos_x,
    double dir_y,
    double dir_x,
    double plane_y,
    double plane_x,
):
    cdef:
        size_t h = texture.shape[0], w = texture.shape[1]
        size_t y, x
        size_t ch, cw, fh, fw
        double ray_dir_y0, ray_dir_y1
        double ray_dir_x0, ray_dir_x1
        int p
        double pos_z, row_distance
        double floor_step_y, floor_step_x
        double floor_y, floor_x
        int cell_y, cell_x, tx, ty

    if draw_ceil:
        ch = ceiling_texture.shape[0]
        cw = ceiling_texture.shape[1]
    if draw_floor:
        fh = floor_texture.shape[0]
        fw = floor_texture.shape[1]

    for y in range(h//2, h):
        ray_dir_y0 = dir_y - plane_y
        ray_dir_y1 = dir_y + plane_y
        ray_dir_x0 = dir_x - plane_x
        ray_dir_x1 = dir_x + plane_x
        p = <int>(y - h // 2)
        pos_z = 0.5 * h
        if p:
            row_distance = pos_z / p
        else:
            row_distance = pos_z
        floor_step_y = row_distance * (ray_dir_y1 - ray_dir_y0) / w
        floor_step_x = row_distance * (ray_dir_x1 - ray_dir_x0) / w
        floor_y = pos_y + row_distance * ray_dir_y0
        floor_x = pos_x + row_distance * ray_dir_x0
        for x in range(w):
            cell_y = <int>floor_y
            cell_x = <int>floor_x

            if draw_ceil:
                ty = <int>(ch * (floor_y - cell_y))
                tx = <int>(cw * (floor_x - cell_x))
                texture[h - y - 1, x, 0] = ceiling_texture[ty, tx, 0]
                texture[h - y - 1, x, 1] = ceiling_texture[ty, tx, 1]
                texture[h - y - 1, x, 2] = ceiling_texture[ty, tx, 2]
                texture[h - y - 1, x, 3] = ceiling_texture[ty, tx, 3]
            if draw_floor:
                ty = <int>(fh * (floor_y - cell_y))
                tx = <int>(fw * (floor_x - cell_x))
                texture[y, x, 0] = floor_texture[ty, tx, 0]
                texture[y, x, 1] = floor_texture[ty, tx, 1]
                texture[y, x, 2] = floor_texture[ty, tx, 2]
                texture[y, x, 3] = floor_texture[ty, tx, 3]

            floor_y += floor_step_y
            floor_x += floor_step_x


@cython.boundscheck(False)
@cython.wraparound(False)
def cast_rays(
    uint8[:, :, ::1] texture,
    ulong[:, ::1] map,
    tuple[double, double] camera_pos,
    double camera_theta,
    double camera_fov,
    list[uint8[:, :, ::1]] wall_textures,
    uint8[:, :, ::1] ceiling_texture,
    uint8[:, :, ::1] floor_texture,
    ulong[::1] sprite_indexes,
    double[:, ::1] sprite_coords,
    list[uint8[:, :, ::1]] sprite_textures,
) -> None:
    cdef int h = <int>texture.shape[0], w = <int>texture.shape[1]
    if not h or not w:
        return

    cdef:
        double pos_y = camera_pos[0], pos_x = camera_pos[1]
        double dir_y = cos(camera_theta), dir_x = sin(camera_theta)
        double plane_y = camera_fov * dir_x, plane_x = camera_fov * -dir_y
        bint draw_ceiling = ceiling_texture is not None
        bint draw_floor = floor_texture is not None

    if draw_ceiling or draw_floor:
        draw_floor_ceiling(
            texture,
            draw_ceiling,
            ceiling_texture,
            draw_floor,
            floor_texture,
            pos_y,
            pos_x,
            dir_y,
            dir_x,
            plane_y,
            plane_x,
        )

    cdef:
        size_t y, x
        size_t mh = map.shape[0], mw = map.shape[1]
        int map_y, map_x
        double camera_x, delta_dist_y, delta_dist_x
        int step_y, step_x
        int line_height
        int initial_y, draw_start_y, draw_end_y
        int tex_h, tex_w, tex_y, tex_x
        double wall_x, step, tex_pos
        bint side, skip
        double perp_wall_dist
        uint8[:, :, ::1] wall_texture
        double *zbuffer = <double*>malloc(w * sizeof(double))

    for x in range(w):
        camera_x = 2 * x / <double>w - 1
        ray_dir_y = dir_y + plane_y * camera_x
        ray_dir_x = dir_x + plane_x * camera_x
        map_y = <int>pos_y
        map_x = <int>pos_x
        delta_dist_y = (
            1e30
            if ray_dir_y == 0
            else fabs(1 / ray_dir_y)
        )
        delta_dist_x = (
            1e30
            if ray_dir_x == 0
            else fabs(1 / ray_dir_x)
        )

        if ray_dir_y < 0:
            step_y = -1
            side_dist_y = (pos_y - map_y) * delta_dist_y
        else:
            step_y = 1
            side_dist_y = (map_y + 1.0 - pos_y) * delta_dist_y

        if ray_dir_x < 0:
            step_x = -1
            side_dist_x = (pos_x - map_x) * delta_dist_x
        else:
            step_x = 1
            side_dist_x = (map_x + 1.0 - pos_x) * delta_dist_x

        # Cast rays.
        skip = 0
        while True:
            if side_dist_x < side_dist_y:
                side_dist_x += delta_dist_x
                map_x += step_x
                side = 0
            else:
                side_dist_y += delta_dist_y
                map_y += step_y
                side = 1

            if map_y < 0 or map_y >= mh or map_x < 0 or map_x >= mw:
                skip = 1  # Outside map.
                break

            if map[map_y, map_x] > 0:
                wall_texture = wall_textures[map[map_y, map_x] - 1]
                tex_h = <int>wall_texture.shape[0]
                tex_w = <int>wall_texture.shape[1]
                if tex_h == 0 or tex_w == 0:
                    skip = 1  # Empty texture?
                break

        if side:
            perp_wall_dist = side_dist_y - delta_dist_y
        else:
            perp_wall_dist = side_dist_x - delta_dist_x
        zbuffer[x] = perp_wall_dist

        if skip or perp_wall_dist == 0:
            continue

        line_height = <int>(h / perp_wall_dist)
        if line_height == 0:
            continue

        initial_y = (h - line_height) // 2
        if initial_y < 0:
            draw_start_y = 0
        else:
            draw_start_y = initial_y
        draw_end_y = (h + line_height + 1) // 2
        if draw_end_y > h:
            draw_end_y = h

        if side:
            wall_x = pos_x + perp_wall_dist * ray_dir_x
        else:
            wall_x = pos_y + perp_wall_dist * ray_dir_y
        wall_x -= floor(wall_x)

        tex_x = <int>(wall_x * <double>tex_w)
        if side == 0 and ray_dir_x > 0 or side == 1 and ray_dir_y < 0:
            tex_x = tex_w - tex_x - 1

        step = <double>tex_h / line_height
        tex_pos = (draw_start_y - initial_y) * step
        for y in range(draw_start_y, draw_end_y):
            tex_y = <int>tex_pos
            tex_pos += step

            if tex_y < 0:
                tex_y = 0
            elif tex_y >= tex_h:
                tex_y = tex_h - 1

            if side:
                # Darken wall.
                texture[y, x, 0] = (wall_texture[tex_y, tex_x, 0] >> 1) & 127
                texture[y, x, 1] = (wall_texture[tex_y, tex_x, 1] >> 1) & 127
                texture[y, x, 2] = (wall_texture[tex_y, tex_x, 2] >> 1) & 127
            else:
                texture[y, x, 0] = wall_texture[tex_y, tex_x, 0]
                texture[y, x, 1] = wall_texture[tex_y, tex_x, 1]
                texture[y, x, 2] = wall_texture[tex_y, tex_x, 2]
            texture[y, x, 3] = wall_texture[tex_y, tex_x, 3]

    cdef:
        size_t nsprites, i
        Sprite *sprites
        Sprite *sprite

    nsprites = sprite_indexes.shape[0]
    sprites = <Sprite*>malloc(nsprites * sizeof(Sprite))
    for i in range(nsprites):
        sprite = &sprites[i]
        sprite.y = sprite_coords[i, 0]
        sprite.x = sprite_coords[i, 1]
        sprite.texture_idx = sprite_indexes[i]
        sprite.distance = (
            (pos_y - sprite.y) * (pos_y - sprite.y)
            + (pos_x - sprite.x) * (pos_x - sprite.x)
        )
    qsort(sprites, nsprites, sizeof(Sprite), &sprite_cmp)

    cdef:
        double inv_det = 1.0 / (plane_x * dir_y - plane_y * dir_x)
        double sprite_y, sprite_x
        double trans_y, trans_x, scale_h, scale_w
        int sprite_screen_x, sprite_h, sprite_w
        int initial_x, draw_start_x, draw_end_x
        uint8[:, :, ::1] sprite_texture

    for i in range(nsprites):
        sprite_y = sprites[i].y - pos_y
        sprite_x = sprites[i].x - pos_x
        trans_y = inv_det * (plane_x * sprite_y - plane_y * sprite_x)
        if trans_y <= 0:
            # Behind the camera.
            continue

        trans_x = inv_det * (dir_y * sprite_x - dir_x * sprite_y)
        sprite_screen_x = <int>((w / 2) * (1 + trans_x / trans_y))

        sprite_h = <int>(h / trans_y)
        if sprite_h == 0:
            continue
        sprite_w = <int>((w / 2) / trans_y)
        if sprite_w == 0:
            continue

        initial_y = h // 2 - sprite_h // 2
        if initial_y < 0:
            draw_start_y = 0
        else:
            draw_start_y = initial_y
        draw_end_y = h // 2 + sprite_h // 2
        if draw_end_y > h:
            draw_end_y = h

        initial_x = sprite_screen_x - sprite_w // 2
        if initial_x < 0:
            draw_start_x = 0
        else:
            draw_start_x = initial_x
        draw_end_x = sprite_screen_x + sprite_w // 2
        if draw_end_x > w:
            draw_end_x = w

        sprite_texture = sprite_textures[sprites[i].texture_idx]
        tex_h = <int>sprite_texture.shape[0]
        tex_w = <int>sprite_texture.shape[1]

        scale_h = tex_h / sprite_h
        scale_w = tex_w / sprite_w

        for x in range(draw_start_x, draw_end_x):
            if trans_y > zbuffer[x]:
                # Sprite behind a wall.
                continue

            tex_x = <int>((x - initial_x) * scale_w)
            if tex_x < 0 or tex_x >= tex_w:
                continue

            for y in range(draw_start_y, draw_end_y):
                tex_y = <int>((y - initial_y) * scale_h)
                if tex_y < 0 or tex_y >= tex_h:
                    continue
                composite(&texture[y, x, 0], &sprite_texture[tex_y, tex_x, 0])

    free(zbuffer)
    free(sprites)


cdef uint8 shade_wall(uint8 value, double distance):
    # ! Should shading be set by kwarg in caster?
    # ! `-0.15` is pretty arbitrary.
    return <uint8>(value * exp(-0.15 * distance))


@cython.boundscheck(False)
@cython.wraparound(False)
def text_cast_rays(
    Cell[:, ::1] canvas,
    ulong[:, ::1] map,
    tuple[double, double] camera_pos,
    double camera_theta,
    double camera_fov,
    list[uint8[:, ::1]] wall_textures,
    ulong[::1] sprite_indexes,
    double[:, ::1] sprite_coords,
    list[Cell[:, ::1]] sprite_textures,
    unsigned int[::1] ascii_map,
) -> None:
    cdef int h = <int>canvas.shape[0], w = <int>canvas.shape[1]
    if not h or not w:
        return

    cdef:
        double pos_y = camera_pos[0], pos_x = camera_pos[1]
        double dir_y = cos(camera_theta), dir_x = sin(camera_theta)
        double plane_y = camera_fov * dir_x, plane_x = camera_fov * -dir_y
        size_t y, x
        size_t mh = map.shape[0], mw = map.shape[1]
        int map_y, map_x
        double camera_x, delta_dist_y, delta_dist_x
        int step_y, step_x
        int line_height
        int initial_y, draw_start_y, draw_end_y
        int tex_h, tex_w, tex_y, tex_x
        double wall_x, step, tex_pos
        bint side, skip
        double perp_wall_dist
        uint8[:, ::1] wall_texture
        double *zbuffer = <double*>malloc(w * sizeof(double))
        uint8 wall_value

    for x in range(w):
        camera_x = 2 * x / <double>w - 1
        ray_dir_y = dir_y + plane_y * camera_x
        ray_dir_x = dir_x + plane_x * camera_x
        map_y = <int>pos_y
        map_x = <int>pos_x
        delta_dist_y = (
            1e30
            if ray_dir_y == 0
            else fabs(1 / ray_dir_y)
        )
        delta_dist_x = (
            1e30
            if ray_dir_x == 0
            else fabs(1 / ray_dir_x)
        )

        if ray_dir_y < 0:
            step_y = -1
            side_dist_y = (pos_y - map_y) * delta_dist_y
        else:
            step_y = 1
            side_dist_y = (map_y + 1.0 - pos_y) * delta_dist_y

        if ray_dir_x < 0:
            step_x = -1
            side_dist_x = (pos_x - map_x) * delta_dist_x
        else:
            step_x = 1
            side_dist_x = (map_x + 1.0 - pos_x) * delta_dist_x

        # Cast rays.
        skip = 0
        while True:
            if side_dist_x < side_dist_y:
                side_dist_x += delta_dist_x
                map_x += step_x
                side = 0
            else:
                side_dist_y += delta_dist_y
                map_y += step_y
                side = 1

            if map_y < 0 or map_y >= mh or map_x < 0 or map_x >= mw:
                skip = 1  # Outside map.
                break

            if map[map_y, map_x] > 0:
                wall_texture = wall_textures[map[map_y, map_x] - 1]
                tex_h = <int>wall_texture.shape[0]
                tex_w = <int>wall_texture.shape[1]
                if tex_h == 0 or tex_w == 0:
                    skip = 1  # Empty texture?
                break

        if side:
            perp_wall_dist = side_dist_y - delta_dist_y
        else:
            perp_wall_dist = side_dist_x - delta_dist_x
        zbuffer[x] = perp_wall_dist

        if skip or perp_wall_dist == 0:
            continue

        line_height = <int>(h / perp_wall_dist)
        if line_height == 0:
            continue

        initial_y = (h - line_height) // 2
        if initial_y < 0:
            draw_start_y = 0
        else:
            draw_start_y = initial_y
        draw_end_y = (h + line_height + 1) // 2
        if draw_end_y > h:
            draw_end_y = h

        if side:
            wall_x = pos_x + perp_wall_dist * ray_dir_x
        else:
            wall_x = pos_y + perp_wall_dist * ray_dir_y
        wall_x -= floor(wall_x)

        tex_x = <int>(wall_x * <double>tex_w)
        if side == 0 and ray_dir_x > 0 or side == 1 and ray_dir_y < 0:
            tex_x = tex_w - tex_x - 1

        step = <double>tex_h / line_height
        tex_pos = (draw_start_y - initial_y) * step
        for y in range(draw_start_y, draw_end_y):
            tex_y = <int>tex_pos
            tex_pos += step

            if tex_y < 0:
                tex_y = 0
            elif tex_y >= tex_h:
                tex_y = tex_h - 1

            # ! How much to shade a side wall? `2.0 * side` for now...
            wall_value = shade_wall(
                wall_texture[tex_y, tex_x], perp_wall_dist + 2.0 * side
            )
            canvas[y, x].ord = ascii_map[wall_value]

    cdef:
        size_t nsprites, i
        Sprite *sprites
        Sprite *sprite

    nsprites = sprite_indexes.shape[0]
    sprites = <Sprite*>malloc(nsprites * sizeof(Sprite))
    for i in range(nsprites):
        sprite = &sprites[i]
        sprite.y = sprite_coords[i, 0]
        sprite.x = sprite_coords[i, 1]
        sprite.texture_idx = sprite_indexes[i]
        sprite.distance = (
            (pos_y - sprite.y) * (pos_y - sprite.y)
            + (pos_x - sprite.x) * (pos_x - sprite.x)
        )
    qsort(sprites, nsprites, sizeof(Sprite), &sprite_cmp)

    cdef:
        double inv_det = 1.0 / (plane_x * dir_y - plane_y * dir_x)
        double sprite_y, sprite_x
        double trans_y, trans_x, scale_h, scale_w
        int sprite_screen_x, sprite_h, sprite_w
        int initial_x, draw_start_x, draw_end_x
        Cell[:, ::1] sprite_texture

    for i in range(nsprites):
        sprite_y = sprites[i].y - pos_y
        sprite_x = sprites[i].x - pos_x
        trans_y = inv_det * (plane_x * sprite_y - plane_y * sprite_x)
        if trans_y <= 0:
            # Behind the camera.
            continue

        trans_x = inv_det * (dir_y * sprite_x - dir_x * sprite_y)
        sprite_screen_x = <int>((w / 2) * (1 + trans_x / trans_y))

        sprite_h = <int>(h / trans_y)
        if sprite_h == 0:
            continue
        sprite_w = <int>((w / 2) / trans_y)
        if sprite_w == 0:
            continue

        initial_y = h // 2 - sprite_h // 2
        if initial_y < 0:
            draw_start_y = 0
        else:
            draw_start_y = initial_y
        draw_end_y = h // 2 + sprite_h // 2
        if draw_end_y > h:
            draw_end_y = h

        initial_x = sprite_screen_x - sprite_w // 2
        if initial_x < 0:
            draw_start_x = 0
        else:
            draw_start_x = initial_x
        draw_end_x = sprite_screen_x + sprite_w // 2
        if draw_end_x > w:
            draw_end_x = w

        sprite_texture = sprite_textures[sprites[i].texture_idx]
        tex_h = <int>sprite_texture.shape[0]
        tex_w = <int>sprite_texture.shape[1]

        scale_h = tex_h / sprite_h
        scale_w = tex_w / sprite_w

        for x in range(draw_start_x, draw_end_x):
            if trans_y > zbuffer[x]:
                # Sprite behind a wall.
                continue

            tex_x = <int>((x - initial_x) * scale_w)
            if tex_x < 0 or tex_x >= tex_w:
                continue

            for y in range(draw_start_y, draw_end_y):
                tex_y = <int>((y - initial_y) * scale_h)
                if tex_y < 0 or tex_y >= tex_h:
                    continue
                if sprite_texture[tex_y, tex_x].ord == 48:
                    continue
                canvas[y, x] = sprite_texture[tex_y, tex_x]

    free(zbuffer)
    free(sprites)



"""A growable string buffer."""
from libc.stdint cimport uint32_t, uint64_t

cdef extern from "_fbuf.h":
    struct fbuf:
        uint64_t size, len
        char *buf

    ssize_t write(ssize_t, const void*, size_t)
    ssize_t fbuf_init(fbuf *f)
    ssize_t fbuf_small_init(fbuf *f)
    void fbuf_free(fbuf *f)
    ssize_t fbuf_grow(fbuf *f, size_t n)
    ssize_t fbuf_put_char(fbuf *f, const char s)
    ssize_t fbuf_putn(fbuf *f, const char *s, size_t len)
    ssize_t fbuf_puts(fbuf *f, const char *s)
    ssize_t fbuf_printf(fbuf *f, const char *fmt, ...)
    ssize_t fbuf_putucs4(fbuf *f, uint32_t wc)
    unsigned int fbuf_equals(fbuf *f, const char *string, size_t len)
    unsigned int fbuf_endswith(fbuf *f, const char *suffix, size_t len)
    ssize_t fbuf_flush_fd(fbuf *f, int fd)
    ssize_t fbuf_read_fd(fbuf *f, int fd, int *size)



"""
Inheritable gadget behaviors.

A `Behavior` is an inheritable class that modifies a gadget.

It should be inherited *before* the base gadget, e.g.,::

    class MovableImage(Movable, Image): ...

Where `Movable` is a `Behavior` and `Image` is the base gadget. In this case,
`MovableImage` is now an `Image` that can be moved around the terminal by clicking and
dragging.
"""

from typing import cast

from ..gadget import Gadget

Behavior = cast(type[Gadget], object)
"""
A `Behavior` is an inheritable class that modifies a gadget.

It should be inherited *before* the base gadget, e.g.,::

    class MovableImage(Movable, Image): ...

Where `Movable` is a `Behavior` and `Image` is the base gadget. In this case,
`MovableImage` is now an `Image` that can be moved around the terminal by clicking and
dragging.
"""



"""Commonly used colors."""

from typing import Final

from .._style import Style
from .color_types import AColor, Color, ColorTheme, SyntaxHighlightTheme

__all__ = [
    "ABLACK",
    "ABLUE",
    "ACYAN",
    "AGREEN",
    "AMAGENTA",
    "ARED",
    "AWHITE",
    "AYELLOW",
    "BLACK",
    "BLUE",
    "CYAN",
    "GREEN",
    "MAGENTA",
    "NEPTUNE_PRIMARY_BG",
    "NEPTUNE_PRIMARY_FG",
    "NEPTUNE_THEME",
    "RED",
    "TRANSPARENT",
    "WHITE",
    "YELLOW",
    "Neptune",
]

WHITE: Final = Color(255, 255, 255)
"""White."""

BLACK: Final = Color(0, 0, 0)
"""Black."""

RED: Final = Color(255, 0, 0)
"""Red."""

GREEN: Final = Color(0, 255, 0)
"""Green."""

BLUE: Final = Color(0, 0, 255)
"""Blue."""

YELLOW: Final = Color(255, 255, 0)
"""Yellow."""

CYAN: Final = Color(0, 255, 255)
"""Cyan."""

MAGENTA: Final = Color(255, 0, 255)
"""Magenta."""

AWHITE: Final = AColor(255, 255, 255)
"""Opaque white."""

ABLACK: Final = AColor(0, 0, 0)
"""Opaque black."""

ARED: Final = AColor(255, 0, 0)
"""Opaque red."""

AGREEN: Final = AColor(0, 255, 0)
"""Opaque green."""

ABLUE: Final = AColor(0, 0, 255)
"""Opaque blue."""

AYELLOW: Final = AColor(255, 255, 0)
"""Opaque yellow."""

ACYAN: Final = AColor(0, 255, 255)
"""Opaque cyan."""

AMAGENTA: Final = AColor(255, 0, 255)
"""Opaque magenta."""

TRANSPARENT: Final = AColor(0, 0, 0, 0)
"""Transparent black."""

NEPTUNE_THEME: Final[ColorTheme] = {
    "primary_fg": "#f6a7a9",
    "primary_bg": "#070c25",
    "text_pad_line_highlight_fg": "#f6a7a9",
    "text_pad_line_highlight_bg": "#0c0e30",
    "text_pad_selection_highlight_fg": "#f6a7a9",
    "text_pad_selection_highlight_bg": "#0f1847",
    "textbox_primary_fg": "#fff0f6",
    "textbox_primary_bg": "#070c25",
    "textbox_selection_highlight_fg": "#fff0f6",
    "textbox_selection_highlight_bg": "#0f1847",
    "textbox_placeholder_fg": "#2a3a92",
    "textbox_placeholder_bg": "#070c25",
    "button_normal_fg": "#dde4ed",
    "button_normal_bg": "#2a3ca0",
    "button_hover_fg": "#fff0f6",
    "button_hover_bg": "#3248c0",
    "button_press_fg": "#fff0f6",
    "button_press_bg": "#c4a219",
    "button_disallowed_fg": "#272b40",
    "button_disallowed_bg": "#070c25",
    "menu_item_hover_fg": "#f2babc",
    "menu_item_hover_bg": "#111834",
    "menu_item_selected_fg": "#ecf3ff",
    "menu_item_selected_bg": "#1b244b",
    "menu_item_disallowed_fg": "#272b40",
    "menu_item_disallowed_bg": "#070c25",
    "titlebar_normal_fg": "#ffe0df",
    "titlebar_normal_bg": "#070c25",
    "titlebar_inactive_fg": "#7d6b71",
    "titlebar_inactive_bg": "#070c25",
    "data_table_sort_indicator_fg": "#ecf3ff",
    "data_table_sort_indicator_bg": "#070c25",
    "data_table_hover_fg": "#f6a7a9",
    "data_table_hover_bg": "#111834",
    "data_table_stripe_fg": "#f6a7a9",
    "data_table_stripe_bg": "#0b1238",
    "data_table_stripe_hover_fg": "#f6a7a9",
    "data_table_stripe_hover_bg": "#0f184a",
    "data_table_selected_fg": "#ecf3ff",
    "data_table_selected_bg": "#111f5e",
    "data_table_selected_hover_fg": "#ecf3ff",
    "data_table_selected_hover_bg": "#1b244b",
    "progress_bar_fg": "#ffe0df",
    "progress_bar_bg": "#2a3ca0",
    "markdown_link_fg": "#376cff",
    "markdown_link_bg": "#070c25",
    "markdown_link_hover_fg": "#4668ff",
    "markdown_link_hover_bg": "#070c25",
    "markdown_inline_code_fg": "#806ae5",
    "markdown_inline_code_bg": "#080b1a",
    "markdown_quote_fg": "#2054e2",
    "markdown_quote_bg": "#0c1b4b",
    "markdown_title_fg": "#cfd1d4",
    "markdown_title_bg": "#292a2d",
    "markdown_image_fg": "#f6a7a9",
    "markdown_image_bg": "#0c1540",
    "markdown_block_code_bg": "#080b1a",
    "markdown_quote_block_code_bg": "#11265d",
    "markdown_header_bg": "#030612",
    "scroll_view_scrollbar": "#070c25",
    "scroll_view_indicator_normal": "#0e1843",
    "scroll_view_indicator_hover": "#111e4f",
    "scroll_view_indicator_press": "#172868",
    "window_border_normal": "#122162",
    "window_border_inactive": "#282c3e",
}
"""Neptune color theme."""

NEPTUNE_PRIMARY_FG: Final = Color.from_hex(NEPTUNE_THEME["primary_fg"])
"""The primary foreground color for the Neptune color theme."""

NEPTUNE_PRIMARY_BG: Final = Color.from_hex(NEPTUNE_THEME["primary_bg"])
"""The primary background color for the Neptune color theme."""

Neptune = SyntaxHighlightTheme(
    default_fg="#f6a7a9",
    default_bg="#070c25",
    cursor_fg="#070c25",
    cursor_bg="#5268e6",
    active_line="#0d0e30",
    selection="#101847",
    highlights={
        "attribute": (Style.NO_STYLE, "#d4864d"),
        "bold": (Style.BOLD, None),
        "comment": (Style.NO_STYLE, "#554492"),
        "comment.documentation": (Style.NO_STYLE, "#4876db"),
        "comment.error": (Style.NO_STYLE, "#d6452b"),
        "comment.note": (Style.NO_STYLE, "#90cf5c"),
        "comment.todo": (Style.NO_STYLE, "#4eb8d8"),
        "comment.warning": (Style.NO_STYLE, "#d37f11"),
        "conditional": (Style.NO_STYLE, "#5bb177"),
        "constant": (Style.BOLD, "#91a5Bd"),
        "constant.builtin": (Style.BOLD, "#17c78c"),
        "constant.builtin.boolean": (Style.BOLD, "#17b865"),
        "constant.character": (Style.BOLD, "#17c78c"),
        "constant.macro": (Style.BOLD, "#7eb0ec"),
        "constructor": (Style.BOLD, "#de9138"),
        "escape": (Style.NO_STYLE, "#17c78c"),
        "error": (Style.NO_STYLE, "#c71726"),
        "exception": (Style.NO_STYLE, "#3db177"),
        "float": (Style.NO_STYLE, "#e47559"),
        "function": (Style.NO_STYLE, "#cb5757"),
        "function.builtin": (Style.BOLD, "#d15656"),
        "function.call": (Style.NO_STYLE, "#cb5757"),
        "function.special": (Style.NO_STYLE, "#d85757"),
        "import": (Style.NO_STYLE, "#6180b3"),
        "include": (Style.NO_STYLE, "#6180b3"),
        "italic": (Style.ITALIC, None),
        "keyword": (Style.NO_STYLE, "#17b865"),
        "keyword.class": (Style.NO_STYLE, "#c75974"),
        "keyword.conditional": (Style.NO_STYLE, "#5bb177"),
        "keyword.directive": (Style.NO_STYLE, "#e9dfcc"),
        "keyword.function": (Style.NO_STYLE, "#c75974"),
        "keyword.modifier": (Style.NO_STYLE, "#b96772"),
        "keyword.operator": (Style.NO_STYLE, "#c76f6f"),
        "keyword.repeat": (Style.NO_STYLE, "#5bb177"),
        "keyword.return": (Style.NO_STYLE, "#5bb177"),
        "keyword.type": (Style.BOLD, "#de9138"),
        "method": (Style.NO_STYLE, "#cb5757"),
        "method.call": (Style.NO_STYLE, "#cb5757"),
        "number": (Style.NO_STYLE, "#e47559"),
        "operator": (Style.NO_STYLE, "#c76f6f"),
        "parameter": (Style.NO_STYLE, "#ce967c"),
        "property": (Style.NO_STYLE, "#fdc0be"),
        "punctuation": (Style.NO_STYLE, "#ce967c"),
        "punctuation.bracket": (Style.NO_STYLE, "#a73e3e"),
        "punctuation.delimiter": (Style.NO_STYLE, "#ce967c"),
        "punctuation.special": (Style.NO_STYLE, "#ce967c"),
        "repeat": (Style.NO_STYLE, "#5bb177"),
        "strikethrough": (Style.STRIKETHROUGH, None),
        "string": (Style.NO_STYLE, "#da7763"),
        "string.documentation": (Style.NO_STYLE, "#4876db"),
        "string.escape": (Style.NO_STYLE, "#17c78c"),
        "string.special.url": (Style.UNDERLINE, None),
        "type": (Style.BOLD, "#de9138"),
        "type.builtin": (Style.BOLD, "#de9138"),
        "type.class": (Style.BOLD, "#de9138"),
        "type.enum": (Style.BOLD, "#a076b9"),
        "variable": (Style.NO_STYLE, "#fdc0be"),
        "variable.builtin": (Style.NO_STYLE, "#c68a6d"),
        "variable.parameter": (Style.NO_STYLE, "#d38a6d"),
    },
)



"""Themable behavior for gadgets."""

from abc import ABC, abstractmethod

from ...colors import NEPTUNE_THEME, Color, ColorTheme, ColorThemeColor, Hexcode
from . import Behavior

__all__ = ["Themable"]


class Themable(ABC, Behavior):
    """
    Themable behavior for a gadget.

    Themable gadgets share a color theme. They must implement :meth:`update_theme`
    which paints the gadget with current theme.

    Whenever the running app's theme is changed, `update_theme` will be called
    for all :class:`Themable` gadgets.

    Methods
    -------
    get_color(color_name)
        Get a color by name from the current color theme.
    update_theme()
        Paint the gadget with current theme.
    """

    color_theme: ColorTheme

    @abstractmethod
    def update_theme(self) -> None:
        """Paint the gadget with current theme."""

    @classmethod
    def get_color(cls, color_name: ColorThemeColor) -> Color:
        """Get a color by name from the current color theme."""
        hexcode: Hexcode
        if color_name not in cls.color_theme:
            if color_name not in NEPTUNE_THEME:
                raise KeyError(f"There is no color {color_name!r}.")
            hexcode = NEPTUNE_THEME[color_name]
        else:
            hexcode = cls.color_theme[color_name]
        return Color.from_hex(hexcode)

    def on_add(self) -> None:
        """Update theme."""
        super().on_add()
        self.update_theme()



"""Easing functions."""

from collections.abc import Callable
from math import cos, pi, sin
from typing import Literal

__all__ = ["EASINGS", "Easing"]

Easing = Literal[
    "linear",
    "in_quad",
    "out_quad",
    "in_out_quad",
    "in_cubic",
    "out_cubic",
    "in_out_cubic",
    "in_quart",
    "out_quart",
    "in_out_quart",
    "in_quint",
    "out_quint",
    "in_out_quint",
    "in_sine",
    "out_sine",
    "in_out_sine",
    "in_exp",
    "out_exp",
    "in_out_exp",
    "in_circ",
    "out_circ",
    "in_out_circ",
    "in_elastic",
    "out_elastic",
    "in_out_elastic",
    "in_back",
    "out_back",
    "in_out_back",
    "in_bounce",
    "out_bounce",
    "in_out_bounce",
]
"""Easings."""


def linear(p: float) -> float:
    """Linear easing."""
    return p


def in_quad(p: float) -> float:
    """In-quad easing."""
    return p**2.0


def out_quad(p: float) -> float:
    """Out-quad easing."""
    return -p * (p - 2.0)


def in_out_quad(p: float) -> float:
    """In-out-quad easing."""
    if p < 0.5:
        return 2.0 * p**2.0

    return -2.0 * p**2.0 + 4.0 * p - 1.0


def in_cubic(p: float) -> float:
    """In-cubic easing."""
    return p**3.0


def out_cubic(p: float) -> float:
    """Out-cubic easing."""
    return (p - 1.0) ** 3.0 + 1.0


def in_out_cubic(p: float) -> float:
    """In-out-cubic easing."""
    if p < 0.5:
        return 4.0 * p**3.0

    return 0.5 * (2.0 * p - 2.0) ** 3.0 + 1.0


def in_quart(p: float) -> float:
    """In-quart easing."""
    return p**4.0


def out_quart(p: float) -> float:
    """Out-quart easing."""
    return 1.0 - (p - 1.0) ** 4.0


def in_out_quart(p: float) -> float:
    """In-out-quart easing."""
    if p < 0.5:
        return 8.0 * p**4.0

    return -8.0 * (p - 1.0) ** 4.0 + 1.0


def in_quint(p: float) -> float:
    """In-quint easing."""
    return p**5.0


def out_quint(p: float) -> float:
    """Out-quint easing."""
    return (p - 1.0) ** 5.0 + 1.0


def in_out_quint(p: float) -> float:
    """In-out-quint easing."""
    if p < 0.5:
        return 16.0 * p**5.0

    return 0.5 * (2.0 * p - 2.0) ** 5.0 + 1.0


def in_sine(p: float) -> float:
    """In-sine easing."""
    return sin((p - 1.0) * pi * 0.5) + 1.0


def out_sine(p: float) -> float:
    """Out-sine easing."""
    return sin(p * pi * 0.5)


def in_out_sine(p: float) -> float:
    """In-out-sine easing."""
    return 0.5 * (1.0 - cos(p * pi))


def in_exp(p: float) -> float:
    """In-exponential easing."""
    if p == 0.0:
        return 0.0

    return 2.0 ** (10.0 * (p - 1.0))


def out_exp(p: float) -> float:
    """Out-exponential easing."""
    if p == 1.0:
        return 1.0

    return 1.0 - 2.0 ** (-10.0 * p)


def in_out_exp(p: float) -> float:
    """In-out-exponential easing."""
    if p == 0.0:
        return 0.0

    if p == 1.0:
        return 1.0

    if p < 0.5:
        return 0.5 * 2.0 ** (20.0 * p - 10.0)

    return -0.5 * 2.0 ** (-20.0 * p + 10.0) + 1.0


def in_circ(p: float) -> float:
    """In-circular easing."""
    return 1.0 - (1.0 - p**2.0) ** 0.5


def out_circ(p: float) -> float:
    """Out-circular easing."""
    p -= 1.0
    return (1.0 - p**2.0) ** 0.5


def in_out_circ(p: float) -> float:
    """In-out-circular easing."""
    p *= 2.0
    if p < 1.0:
        return -0.5 * (1.0 - p**2.0) ** 0.5 + 0.5

    p -= 2.0
    return 0.5 * (1.0 - p**2.0) ** 0.5 + 0.5


def in_elastic(p: float) -> float:
    """In-elastic easing."""
    return sin(6.5 * pi * p) * 2.0 ** (10.0 * (p - 1.0))


def out_elastic(p: float) -> float:
    """Out-elastic easing."""
    return sin(-6.5 * pi * (p + 1.0)) * 2.0 ** (-10.0 * p) + 1.0


def in_out_elastic(p: float) -> float:
    """In-out-elastic easing."""
    if p < 0.5:
        return 0.5 * sin(13.0 * pi * p) * 2.0 ** (20.0 * p - 10.0)

    return 0.5 * (sin(-13.0 * pi * p) * 2.0 ** (-20.0 * p + 10.0) + 2.0)


def in_back(p: float) -> float:
    """In-back easing."""
    return p**2.0 * (2.70158 * p - 1.70158)


def out_back(p: float) -> float:
    """Out-back easing."""
    p -= 1.0
    return p**2.0 * (2.70158 * p + 1.70158) + 1.0


def in_out_back(p: float) -> float:
    """In-out-back easing."""
    p *= 2.0
    if p < 1.0:
        return 0.5 * (p**2.0 * (3.5949095 * p - 2.5949095))

    p -= 2.0
    return 0.5 * (p**2.0 * (3.5949095 * p + 2.5949095) + 2.0)


def out_bounce(p: float) -> float:
    """Out-bounce easing."""
    if p < 4.0 / 11.0:
        return 121.0 * p**2.0 / 16.0

    if p < 8.0 / 11.0:
        return 363.0 / 40.0 * p**2.0 - 99.0 / 10.0 * p + 17.0 / 5.0

    if p < 9.0 / 10.0:
        return 4356.0 / 361.0 * p**2.0 - 35442.0 / 1805.0 * p + 16061.0 / 1805.0

    return 54.0 / 5.0 * p**2.0 - 513.0 / 25.0 * p + 268.0 / 25.0


def in_bounce(p: float) -> float:
    """In-bounce easing."""
    return 1.0 - out_bounce(1.0 - p)


def in_out_bounce(p: float) -> float:
    """In-out-bounce easing."""
    if p < 0.5:
        return 0.5 * in_bounce(2.0 * p)

    return 0.5 * out_bounce(2.0 * p - 1.0) + 0.5


EASINGS: dict[Easing, Callable[[float], float]] = {
    "linear": linear,
    "in_quad": in_quad,
    "out_quad": out_quad,
    "in_out_quad": in_out_quad,
    "in_cubic": in_cubic,
    "out_cubic": out_cubic,
    "in_out_cubic": in_out_cubic,
    "in_quart": in_quart,
    "out_quart": out_quart,
    "in_out_quart": in_out_quart,
    "in_quint": in_quint,
    "out_quint": out_quint,
    "in_out_quint": in_out_quint,
    "in_sine": in_sine,
    "out_sine": out_sine,
    "in_out_sine": in_out_sine,
    "in_exp": in_exp,
    "out_exp": out_exp,
    "in_out_exp": in_out_exp,
    "in_circ": in_circ,
    "out_circ": out_circ,
    "in_out_circ": in_out_circ,
    "in_elastic": in_elastic,
    "out_elastic": out_elastic,
    "in_out_elastic": in_out_elastic,
    "in_back": in_back,
    "out_back": out_back,
    "in_out_back": in_out_back,
    "in_bounce": in_bounce,
    "out_bounce": out_bounce,
    "in_out_bounce": in_out_bounce,
}



"""An ans (ANSI art) file viewer."""

from __future__ import annotations

import re
from pathlib import Path
from typing import Final, cast

import numpy as np

from ..colors import BLACK, WHITE, Color, lerp_colors
from ..geometry import Point, Size
from ..text_tools import Cell0D, Cell2D, new_cell
from .gadget import Gadget, PosHint, SizeHint
from .scroll_view import ScrollView
from .text import Pointlike, Sizelike, Text

_CSI_RE: Final = re.compile(r"\x1b\[[;\d]*\w")
"""Control sequence pattern."""
# Colors from VGA column of: https://en.wikipedia.org/wiki/ANSI_escape_code#3-bit_and_4-bit
_LOW_INTENSITY: Final = (
    BLACK,
    Color(170, 0, 0),
    Color(0, 170, 0),
    Color(170, 85, 0),
    Color(0, 0, 170),
    Color(170, 0, 170),
    Color(0, 170, 170),
    Color(170, 170, 170),
)
"""Low intensity terminal colors."""
_HIGH_INTENSITY: Final = (
    Color(85, 85, 85),
    Color(255, 85, 85),
    Color(85, 255, 85),
    Color(255, 255, 85),
    Color(85, 85, 255),
    Color(255, 85, 255),
    Color(85, 255, 255),
    WHITE,
)
"""High intensity terminal colors."""
_UTF8_CHARS: Final = (
    " ☺☻♥♦♣♠•◘○◙♂♀♪♫☼"
    "►◄↕‼¶§▬↨↑↓→←∟↔▲▼"
    " !\"#$%&'()*+,-./"
    "0123456789:;<=>?"
    "@ABCDEFGHIJKLMNO"
    "PQRSTUVWXYZ[\\]^_"
    "`abcdefghijklmno"
    "pqrstuvwxyz{|}~⌂"
    "ÇüéâäàåçêëèïîìÄÅ"
    "ÉæÆôöòûùÿÖÜ¢£¥₧ƒ"
    "áíóúñÑªº¿⌐¬½¼¡«»"
    "░▒▓│┤╡╢╖╕╣║╗╝╜╛┐"
    "└┴┬├─┼╞╟╚╔╩╦╠═╬╧"
    "╨╤╥╙╘╒╓╫╪┘┌█▄▌▐▀"
    "αßΓπΣσµτΦΘΩδ∞φε∩"
    "≡±≥≤⌠⌡÷≈°∙·√ⁿ²■ "
)  # See: https://wikipedia.org/wiki/Code_page_437
_CP437_TO_UTF8: Final = {i: char for i, char in enumerate(_UTF8_CHARS)}
"""Translation table from code page 437 to UTF-8."""
# Following ords interpreted as escapes:
del _CP437_TO_UTF8[0x0A], _CP437_TO_UTF8[0x1A], _CP437_TO_UTF8[0x1B]


class _AnsReader:
    def __init__(self, path: Path, width: int = 80, guess_width: bool = True):
        self.width: int = width
        """Width of ansi art. Determines when to wrap characters."""
        self.guess_width = guess_width
        """Whether to guess width."""
        self.data: str = path.read_text("cp437").translate(_CP437_TO_UTF8)
        """Ansi data."""

        footer_index = self.data[::-1].find("\x1a")
        if footer_index != -1:
            self.data = self.data[: -footer_index - 1]

    def _new_line(self):
        self.lines.append(
            [
                new_cell(fg_color=_LOW_INTENSITY[7], bg_color=_LOW_INTENSITY[0])
                for _ in range(self.width)
            ]
        )

    def _add_chars(self, start: int, stop: int) -> None:
        if isinstance(self.fg_color, int):
            if self.high_intensity:
                fg_color = _HIGH_INTENSITY[self.fg_color]
            else:
                fg_color = _LOW_INTENSITY[self.fg_color]
        else:
            fg_color = self.fg_color

        if isinstance(self.bg_color, int):
            bg_color = _LOW_INTENSITY[self.bg_color]
        else:
            bg_color = self.bg_color

        for i in range(start, stop):
            char = self.data[i]

            if char == "\n":
                self.cursor_y += 1
                self.cursor_x = 0
            else:
                while self.cursor_y >= len(self.lines):
                    self._new_line()

                self.lines[self.cursor_y][self.cursor_x] = new_cell(
                    ord(char), fg_color=fg_color, bg_color=bg_color
                )

                self.cursor_x += 1
                if self.cursor_x >= self.width:
                    self.cursor_y += 1
                    self.cursor_x = 0

    def _parse_ansi(self) -> None:
        current_index = 0
        for match in _CSI_RE.finditer(self.data):
            start, stop = match.span()
            self._add_chars(current_index, start)
            current_index = stop

            escape = match[0]
            command = escape[-1]
            params = (int(n) for n in re.findall(r"\d+", escape))
            if command == "m":
                for param in params:
                    if param == 0:
                        self.high_intensity = False
                        self.fg_color = 7
                        self.bg_color = 0
                    elif param == 1:
                        self.high_intensity = True
                    elif 30 <= param < 38:
                        self.fg_color = param - 30
                    elif 40 <= param < 48:
                        self.bg_color = param - 40
                    elif param == 38 or param == 39:
                        mode = next(params, -1)
                        if mode == 2:
                            color = Color(*(next(params, 0) for _ in range(3)))
                        elif mode == 5:
                            n = next(params, 0)
                            if 0 <= n < 8:
                                color = n
                            elif n < 16:
                                self.high_intensity = True
                                color = n - 8
                            elif n < 232:
                                remainder, b = divmod(n - 16, 6)
                                r, g = divmod(remainder, 6)
                                color = Color(r * 55, g * 55, b * 55)
                            elif n < 256:
                                color = lerp_colors(BLACK, WHITE, (n - 232) / 23)
                            else:
                                continue
                        else:
                            continue

                        if param == 38:
                            self.fg_color = color
                        else:
                            self.bg_color = color
            elif command == "t":
                try:
                    mode, r, g, b = params
                except ValueError:
                    continue

                if mode == 0:
                    self.bg_color = Color(r, g, b)
                elif mode == 1:
                    self.fg_color = Color(r, g, b)
            elif command == "C":
                self.cursor_x = self.cursor_x + next(params, 1)
            elif command == "A":
                self.cursor_y = max(self.cursor_y - next(params, 1), 0)
            elif command == "B":
                self.cursor_y += next(params, 1)
            elif command == "D":
                self.cursor_x = max(self.cursor_x - next(params, 1), 0)
            elif command == "E":
                self.cursor_y += next(params, 1)
                self.cursor_x = 0
            elif command == "F":
                self.cursor_y = max(self.cursor_y - next(params, 1), 0)
                self.cursor_x = 0
            elif command == "G":
                self.cursor_x = self.width - 1
            elif command == "H":
                self.cursor_y = next(params, 1) - 1
                self.cursor_x = next(params, 1) - 1

        self._add_chars(current_index, len(self.data))

    def _create_canvas(self) -> Cell2D:
        canvas = np.full(
            (len(self.lines), self.width),
            new_cell(fg_color=_LOW_INTENSITY[7], bg_color=_LOW_INTENSITY[0]),
        )
        for y, line in enumerate(self.lines):
            for x, cell in enumerate(line):
                canvas[y, x] = cell
        return canvas

    def _guess_read(self) -> Cell2D:
        self.high_intensity = False
        self.fg_color = 7
        self.bg_color = 0

        self.lines: list[list[Cell0D]] = []
        self.cursor_y: int = 0
        self.cursor_x: int = 0
        try:
            self._parse_ansi()
            return self._create_canvas()
        finally:
            del self.lines, self.cursor_y, self.cursor_x

    def read(self) -> Cell2D | None:
        # How to determine the width of an ans file?
        # Most are 80 width, luckily, but there are exceptions! Sometimes there is a way
        # to discover an incorrect width assumption: an index error. An index error
        # implies cursor was moved to end of a line and then characters were written
        # which implies the line should've been longer which implies the width is
        # incorrect. In this case, try again with increased width.
        if self.guess_width:
            for _ in range(40):
                try:
                    return self._guess_read()
                except IndexError:
                    self.width += 1
        else:
            return self._guess_read()


class AnsViewer(Gadget):
    r"""
    An ans (ANSI art) file viewer.

    Parameters
    ----------
    path : Path
        Path to ans file.
    ans_width : int
        Column width of ansi art.
    guess_width : bool
        Whether to guess column width if ``ans_width`` results in an error.
    size : Sizelike, default: Size(10, 10)
        Size of gadget.
    pos : Pointlike, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    ans_width : int
        Column width of ansi art.
    guess_width : bool
        Whether to guess column width if ``ans_width`` results in an error.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : TotalSizeHint
        Size as a proportion of parent's height and width.
    pos_hint : TotalPosHint
        Position as a proportion of parent's height and width.
    parent : Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App | None
        The running app.

    Methods
    -------
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(gadget_it, \*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        path: Path,
        ans_width: int = 80,
        guess_width: bool = True,
        size: Sizelike = Size(10, 10),
        pos: Pointlike = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        super().__init__(
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        self._scroll_view = ScrollView(
            size_hint={"height_hint": 1.0, "width_hint": 1.0}, dynamic_bars=True
        )
        self.add_gadget(self._scroll_view)

        self._ans_reader = _AnsReader(path, ans_width, guess_width)
        self._read_ans()

    @property
    def ans_width(self) -> int:
        """Column width of ansi art."""
        return self._ans_reader.width

    @ans_width.setter
    def ans_width(self, ans_width: int):
        self._ans_reader.width = ans_width
        self._read_ans()

    @property
    def guess_width(self) -> bool:
        """Whether to guess column width if ``ans_width`` results in an error."""
        return self._ans_reader.guess_width

    @guess_width.setter
    def guess_width(self, guess_width: bool):
        self._ans_reader.guess_width = guess_width

    def _read_ans(self):
        canvas = self._ans_reader.read()
        if canvas is None:
            return
        ans = Text(size=cast(Sizelike, canvas.shape))
        ans.canvas = canvas
        self._scroll_view.view = ans
        self.on_size()



"""A linux VT100 terminal."""

import asyncio
import signal
import termios
import tty
from collections.abc import Callable

from .events import Event, ResizeEvent
from .vt100_terminal import Vt100Terminal


class LinuxTerminal(Vt100Terminal):
    """
    A linux VT100 terminal.

    Attributes
    ----------
    stdin: int
        The stdin file descriptor.
    stdout: int
        The stdout file descriptor.
    in_alternate_screen : bool
        Whether the alternate screen buffer is enabled.

    Methods
    -------
    process_stdin()
        Read from stdin and feed data into input parser to generate events.
    raw_mode()
        Set terminal to raw mode.
    restore_console()
        Restore console to its original mode.
    attach(event_handler)
        Start generating events from stdin.
    unattach()
        Stop generating events from stdin.
    feed(input_, reset_before)
        Write bytes to stdin parser and return generated events.
    events()
        Return a list of input events and reset the event buffer.
    get_size()
        Get terminal size.
    write(out)
        Write bytes directly to the out-buffer.
    flush()
        Write out-buffer to output stream and flush.
    set_title(title)
        Set terminal title.
    enter_alternate_screen()
        Enter alternate screen buffer.
    exit_alternate_screen()
        Exit alternate screen buffer.
    enable_mouse_support()
        Enable mouse support in terminal.
    disable_mouse_support()
        Disable mouse support in terminal.
    enable_sgr_pixels()
        Enable SGR-PIXELS mouse mode.
    reset_attributes()
        Reset character attributes.
    enable_bracketed_paste()
        Enable bracketed paste in terminal.
    disable_bracketed_paste()
        Disable bracketed paste in terminal.
    show_cursor()
        Show cursor in terminal.
    hide_cursor()
        Hide cursor in terminal.
    enable_reporting_focus()
        Enable reporting terminal focus.
    disable_reporting_focus()
        Disable reporting terminal focus.
    request_cursor_position_report()
        Report current cursor position.
    request_foreground_color()
        Report terminal foreground color.
    request_background_color()
        Report terminal background color.
    request_device_attributes()
        Report device attributes.
    request_pixel_geometry()
        Report pixel geometry per cell.
    request_terminal_geometry()
        Report pixel geometry of terminal.
    request_sgr_pixels_supported()
        Report whether SGR-PIXELS mouse mode is supported.
    request_synchronized_update_mode_supported()
        Report whether synchronized update mode is supported.
    expect_dsr()
        Return whether a device status report is expected.
    move_cursor(pos)
        Move cursor to ``pos``.
    erase_in_display(n)
        Clear part of the screen.
    """

    def raw_mode(self) -> None:
        """Set terminal to raw mode."""
        self._original_mode = termios.tcgetattr(self.stdin)
        attrs_raw = termios.tcgetattr(self.stdin)
        attrs_raw[tty.LFLAG] &= ~(
            termios.ECHO | termios.ICANON | termios.IEXTEN | termios.ISIG
        )
        attrs_raw[tty.IFLAG] &= ~(
            termios.IXON | termios.IXOFF | termios.ICRNL | termios.INLCR | termios.IGNCR
        )
        attrs_raw[tty.CC][termios.VMIN] = 1
        termios.tcsetattr(self.stdin, termios.TCSANOW, attrs_raw)

    def restore_console(self) -> None:
        """Restore console to its original mode."""
        termios.tcsetattr(self.stdin, termios.TCSANOW, self._original_mode)
        del self._original_mode

    def attach(self, event_handler: Callable[[list[Event]], None]) -> None:
        """
        Start generating events from stdin.

        Parameters
        ----------
        event_handler : Callable[[list[Event]], None]
            Callable that handles input events.
        """
        self._event_buffer.clear()
        self._event_handler = event_handler

        def process():
            self.process_stdin()
            if self._event_handler is not None:
                self._event_handler(self.events())

        loop = asyncio.get_running_loop()
        loop.add_reader(self.stdin, process)

        def on_resize(*_):
            self._event_buffer.append(ResizeEvent(self.get_size()))
            loop.call_soon_threadsafe(process)

        signal.signal(signal.SIGWINCH, on_resize)

    def unattach(self) -> None:
        """Stop generating events from stdin."""
        self._event_handler = None
        loop = asyncio.get_running_loop()
        loop.remove_reader(self.stdin)
        signal.signal(signal.SIGWINCH, signal.SIG_DFL)



"""A spotlight effect."""

import asyncio
from dataclasses import dataclass
from random import randrange
from typing import cast

import numpy as np

from ...array_types import RGB_2D
from ...colors import BLUE, GREEN, RED, Color
from ...geometry import BezierCurve, Pointlike, Size, move_along_path
from ..text import Text


@dataclass
class _SpotLight:
    pos: Pointlike
    color: Color
    radius: float = 8
    falloff: float = 0.3


async def spotlights_effect(text: Text):
    """
    Create a spotlight effect.

    Parameters
    ----------
    text : Text
        The text gadget which to run the effect.

    Warnings
    --------
    Modifying `text` size while effect is running will break the effect.
    """
    cover = Text(size=text.size)
    cover.canvas[:] = text.canvas

    in_fg = cast(RGB_2D, text.canvas["fg_color"])
    in_bg = cast(RGB_2D, text.canvas["bg_color"])
    out_fg = cast(RGB_2D, cover.canvas["fg_color"])
    out_bg = cast(RGB_2D, cover.canvas["bg_color"])

    spotlights = [
        _SpotLight(pos=_random_point(text.size), color=RED),
        _SpotLight(pos=_random_point(text.size), color=GREEN),
        _SpotLight(pos=_random_point(text.size), color=BLUE),
    ]

    text.add_gadget(cover)
    for _ in range(5):
        positions = [_random_point(text.size) for _ in spotlights]
        await _move_spotlights(in_fg, in_bg, out_fg, out_bg, spotlights, positions)

    center = text.height // 2, text.width // 2
    positions = [center for _ in spotlights]
    await _move_spotlights(in_fg, in_bg, out_fg, out_bg, spotlights, positions)
    await _grow_spotlights(in_fg, in_bg, out_fg, out_bg, spotlights)
    text.remove_gadget(cover)


def _random_point(size: Size):
    h, w = size
    return randrange(h), randrange(w)


def _draw_spotlights(
    in_fg: RGB_2D,
    in_bg: RGB_2D,
    out_fg: RGB_2D,
    out_bg: RGB_2D,
    spotlights: list[_SpotLight],
):
    out_fg[:] = in_fg
    out_bg[:] = in_bg
    weighted = np.zeros_like(in_fg, float)
    for spotlight in spotlights:
        h, w, _ = in_fg.shape
        y, x = spotlight.pos
        x /= 2
        ys, xs = np.indices((h, w)).astype(float)
        xs /= 2

        distances = ((ys - y) ** 2 + (xs - x) ** 2) ** 0.5
        distances[distances > spotlight.radius] = np.inf
        weights = np.exp(-spotlight.falloff * distances)
        color = np.array(spotlight.color, float)
        color /= 255
        weighted += color * weights[:, :, None]

    weighted.clip(0, 1, out=weighted)
    out_fg[:] = (in_fg * weighted).astype(np.uint8)
    out_bg[:] = (in_bg * weighted).astype(np.uint8)


async def _move_spotlights(
    in_fg: RGB_2D,
    in_bg: RGB_2D,
    out_fg: RGB_2D,
    out_bg: RGB_2D,
    spotlights: list[_SpotLight],
    positions: list[Pointlike],
):
    def draw(*args):
        _draw_spotlights(in_fg, in_bg, out_fg, out_bg, spotlights)

    paths = []
    for spotlight, position in zip(spotlights, positions):
        curve = BezierCurve(np.array([spotlight.pos, position], float))
        paths.append([curve])

    motions = [
        move_along_path(
            spotlight,
            path,
            speed=20,
            easing="in_out_quad",
            on_start=draw,
            on_progress=draw,
            on_complete=draw,
        )
        for spotlight, path in zip(spotlights, paths)
    ]
    await asyncio.gather(*motions)


async def _grow_spotlights(
    in_fg: RGB_2D,
    in_bg: RGB_2D,
    out_fg: RGB_2D,
    out_bg: RGB_2D,
    spotlights: list[_SpotLight],
):
    h, w, _ = in_fg.shape

    end_radius = ((h / 2) ** 2 + (w / 2) ** 2) ** 0.5
    frames = 100

    all_radii = [
        np.linspace(spotlight.radius, end_radius, frames) for spotlight in spotlights
    ]
    all_falloffs = [
        np.linspace(spotlight.falloff, 0, frames) for spotlight in spotlights
    ]

    for i in range(frames):
        for spotlight, radii, falloffs in zip(spotlights, all_radii, all_falloffs):
            spotlight.radius = radii[i]
            spotlight.falloff = falloffs[i]

        _draw_spotlights(in_fg, in_bg, out_fg, out_bg, spotlights)
        await asyncio.sleep(0.02)



"""Terminal Events."""

from dataclasses import dataclass
from typing import Literal

from ..colors import Color
from ..geometry import Point, Size

__all__ = [
    "CharKey",
    "ColorReportEvent",
    "CursorPositionReportEvent",
    "DeviceAttributesReportEvent",
    "Event",
    "FocusEvent",
    "Key",
    "KeyEvent",
    "MouseButton",
    "MouseEvent",
    "MouseEventType",
    "PasteEvent",
    "PixelGeometryReportEvent",
    "ResizeEvent",
    "SpecialKey",
]

# fmt: off
SpecialKey = Literal[
    "backspace", "delete", "down", "end", "enter", "escape", "f1", "f2", "f3", "f4",
    "f5", "f6", "f7", "f8", "f9", "f10", "f11", "f12", "f13", "f14", "f15", "f16",
    "f17", "f18", "f19", "f20", "f21", "f22", "f23", "f24", "home", "insert", "left",
    "page_down", "page_up", "right", "scroll_down", "scroll_up", "tab", "up",
]
"""A special keyboard key."""
CharKey = Literal[
    " ", "!", '"', "#", "$", "%", "&", "'", "(", ")", "*", "+", ",", "-", ".", "/", "0",
    "1", "2", "3", "4", "5", "6", "7", "8", "9", ":", ";", "<", "=", ">", "?", "@", "A",
    "B", "C", "D", "E", "F", "G", "H", "I", "J", "K", "L", "M", "N", "O", "P", "Q", "R",
    "S", "T", "U", "V", "W", "X", "Y", "Z", "[", "\\", "]", "^", "_", "`", "a", "b",
    "c", "d", "e", "f", "g", "h", "i", "j", "k", "l", "m", "n", "o", "p", "q", "r", "s",
    "t", "u", "v", "w", "x", "y", "z", "{", "|", "}", "~",
]
"""A printable character keyboard key."""
# fmt: on
Key = SpecialKey | CharKey
"""A keyboard key."""
MouseButton = Literal["left", "middle", "no_button", "right"]
"""A mouse button."""
MouseEventType = Literal[
    "mouse_down", "mouse_move", "mouse_up", "scroll_down", "scroll_up"
]
"""A mouse event type."""


class Event:
    """Base event."""


@dataclass
class UnknownEscapeSequence(Event):
    """
    Event generated from an unknown ansi escape sequence.

    Parameters
    ----------
    escape : bytes
        The unknown ansi escape sequence.

    Attributes
    ----------
    escape : bytes
        The unknown ansi escape sequence.
    """

    escape: bytes
    """The unknown ansi escape sequence."""


@dataclass
class ResizeEvent(Event):
    """
    A terminal resize event.

    Parameters
    ----------
    size : Size
        The new terminal size.

    Attributes
    ----------
    size : Size
        The new terminal size.
    """

    size: Size
    """The new terminal size."""


@dataclass
class CursorPositionReportEvent(Event):
    """
    A cursor position report event.

    Parameters
    ----------
    pos : Point
        The reported cursor position.

    Attributes
    ----------
    pos : Point
        The reported cursor position.
    """

    pos: Point
    """The reported cursor position."""


@dataclass
class ColorReportEvent(Event):
    """
    A background or foreground color report event.

    Parameters
    ----------
    kind : Literal["fg", "bg"]
        Whether report is for a foreground ("fg") or background ("bg") color.
    color : Color
        The reported color.

    Attributes
    ----------
    kind : Literal["fg", "bg"]
        Whether report is for a foreground ("fg") or background ("bg") color.
    color : Color
        The reported color.
    """

    kind: Literal["fg", "bg"]
    """Whether report is for a foreground ("fg") or background ("bg") color."""
    color: Color
    """The reported color."""


@dataclass
class DeviceAttributesReportEvent(Event):
    """
    A primary device attributes report event.

    A description of the primary device attributes can be found at https://vt100.net/docs/vt510-rm/chapter4.html.
    (Search text for "Primary Device Attributes").

    Parameters
    ----------
    device_attributes : frozenset[int]
        Reported terminal attributes.

    Attributes
    ----------
    device_attributes : frozenset[int]
        Reported terminal attributes.
    """

    device_attributes: frozenset[int]
    """Reported terminal attributes."""


@dataclass
class PixelGeometryReportEvent(Event):
    """
    A pixel geometry report.

    Parameters
    ----------
    kind : Literal["cell", "terminal"]
        Whether report is pixels per cell or pixels in terminal.
    geometry : Size
        Size of the terminal cells or terminal (depending on ``kind``) in pixels.

    Attributes
    ----------
    kind : Literal["cell", "terminal"]
        Whether report is pixels per cell or pixels in terminal.
    geometry : Size
        Size of the terminal cells or terminal (depending on ``kind``) in pixels.
    """

    kind: Literal["cell", "terminal"]
    """Whether report is pixels per cell or pixels in terminal."""
    geometry: Size
    """Size of the terminal cells or terminal (depending on ``kind``) in pixels."""


@dataclass
class KeyEvent(Event):
    """
    A key event.

    Parameters
    ----------
    key : Key
        The pressed key.
    alt : bool, default: False
        Whether alt was pressed.
    ctrl : bool, default: False
        Whether ctrl was pressed.
    shift : bool, default: False
        Whether shift was pressed.

    Attributes
    ----------
    key : Key
        The pressed key.
    alt : bool
        Whether alt was pressed.
    ctrl : bool
        Whether ctrl was pressed.
    shift : bool
        Whether shift was pressed.
    meta : bool
        Alias for ``alt``.
    control : bool
        Alias for ``ctrl``.
    """

    key: Key
    """The pressed key."""
    alt: bool = False
    """Whether alt was pressed."""
    ctrl: bool = False
    """Whether ctrl was pressed."""
    shift: bool = False
    """Whether shift was pressed."""

    @property
    def meta(self) -> bool:
        """Alias for ``alt``."""
        return self.alt

    @meta.setter
    def meta(self, meta: bool):
        self.alt = meta

    @property
    def control(self) -> bool:
        """Alias for ``ctrl``."""
        return self.ctrl

    @control.setter
    def control(self, control: bool):
        self.ctrl = control


@dataclass
class MouseEvent(Event):
    """
    A mouse event.

    Parameters
    ----------
    pos : Point
        The mouse position.
    button : MouseButton
        The mouse button.
    event_type : MouseEventType
        The mouse event type.
    alt : bool
        Whether alt was pressed.
    ctrl : bool
        Whether ctrl was pressed.
    shift : bool
        Whether shift was pressed.
    dy : int
        The change in y-coordinate of the mouse position.
    dx : int
        The change in x-coordinate of the mouse position.
    nclicks : int, default: 0
        The number of consecutive ``"mouse_down"`` events with same button.
    pixels_pos : Point | None
        The mouse position in pixels.

    Attributes
    ----------
    pos : Point
        The mouse position.
    button : MouseButton
        The mouse button.
    event_type : MouseEventType
        The mouse event type.
    alt : bool
        Whether alt was pressed.
    ctrl : bool
        Whether ctrl was pressed.
    shift : bool
        Whether shift was pressed.
    dy : int
        The change in y-coordinate of the mouse position.
    dx : int
        The change in x-coordinate of the mouse position.
    nclicks : int
        The number of consecutive ``"mouse_down"`` events with same button.
    pixels_pos : Point | None
        The mouse position in pixels.
    meta : bool
        Alias for ``alt``.
    control : bool
        Alias for ``ctrl``.
    """

    pos: Point
    """The mouse position."""
    button: MouseButton
    """The mouse button."""
    event_type: MouseEventType
    """The mouse event type."""
    alt: bool
    """Whether alt was pressed."""
    ctrl: bool
    """Whether ctrl was pressed."""
    shift: bool
    """Whether shift was pressed."""
    dy: int
    """The change in y-coordinate of the mouse position."""
    dx: int
    """The change in x-coordinate of the mouse position."""
    nclicks: int = 0
    """The number of consecutive ``"mouse_down"`` events with same button."""
    pixels_pos: Point | None = None
    """The mouse position in pixels."""

    @property
    def meta(self) -> bool:
        """Alias for ``alt``."""
        return self.alt

    @meta.setter
    def meta(self, meta: bool):
        self.alt = meta

    @property
    def control(self) -> bool:
        """Alias for ``ctrl``."""
        return self.ctrl

    @control.setter
    def control(self, control: bool):
        self.ctrl = control


@dataclass
class PasteEvent(Event):
    """
    A paste event.

    Parameters
    ----------
    paste : str
        The paste content.

    Attributes
    ----------
    paste : str
        The paste content.
    """

    paste: str
    """The paste content."""


@dataclass
class FocusEvent(Event):
    """
    A focus event.

    Parameters
    ----------
    focus : Literal["in", "out"]
        The type of focus; either ``"in"`` or ``"out"``.

    Attributes
    ----------
    focus : Literal["in", "out"]
        The type of focus; either ``"in"`` or ``"out"``.
    """

    focus: Literal["in", "out"]
    """The type of focus; either ``"in"`` or ``"out"``."""


@dataclass
class DECReplyModeEvent(Event):
    """
    A DEC reply event.

    Parameters
    ----------
    mode : int
        The DEC mode.
    value : Literal[0, 1, 2, 3, 4]
        The value of the mode.

    Attributes
    ----------
    mode : int
        The DEC mode.
    value : Literal[0, 1, 2, 3, 4]
        The value of the mode.
    """

    mode: int
    """The DEC mode."""
    value: Literal[0, 1, 2, 3, 4]
    """
    The value of the mode.

    0 indicates mode is not recognized.
    1 indicates mode is on.
    2 indicates mode is off.
    3 indicates mode is permanently on.
    4 indicates mode is permanently off.
    """
